'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any, Union, Tuple, List, Set

from pm4py.objects.conversion.log import converter
from pm4py.objects.log.obj import EventLog, Trace, Event
from pm4py.objects.log.util import dataframe_utils
from pm4py.util import constants, pandas_utils
from pm4py.util import exec_utils
from pm4py.util import xes_constants as xes
from pm4py.util import xes_constants
from pm4py.utils import is_polars_lazyframe


class Parameters(Enum):
    ENABLE_ACTIVITY_DEF_REPRESENTATION = "enable_activity_def_representation"
    ENABLE_SUCC_DEF_REPRESENTATION = "enable_succ_def_representation"
    STR_TRACE_ATTRIBUTES = "str_tr_attr"
    STR_EVENT_ATTRIBUTES = "str_ev_attr"
    NUM_TRACE_ATTRIBUTES = "num_tr_attr"
    NUM_EVENT_ATTRIBUTES = "num_ev_attr"
    STR_EVSUCC_ATTRIBUTES = "str_evsucc_attr"
    FEATURE_NAMES = "feature_names"
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    RESOURCE_KEY = constants.PARAMETER_CONSTANT_RESOURCE_KEY
    EPSILON = "epsilon"
    DEFAULT_NOT_PRESENT = "default_not_present"
    ENABLE_ALL_EXTRA_FEATURES = "enable_all_extra_features"
    ENABLE_CASE_DURATION = "enable_case_duration"
    ADD_CASE_IDENTIFIER_COLUMN = "add_case_identifier_column"
    ENABLE_TIMES_FROM_FIRST_OCCURRENCE = "enable_times_from_first_occurrence"
    ENABLE_TIMES_FROM_LAST_OCCURRENCE = "enable_times_from_last_occurrence"
    ENABLE_DIRECT_PATHS_TIMES_LAST_OCC = "enable_direct_paths_times_last_occ"
    ENABLE_INDIRECT_PATHS_TIMES_LAST_OCC = (
        "enable_indirect_paths_times_last_occ"
    )
    ENABLE_WORK_IN_PROGRESS = "enable_work_in_progress"
    ENABLE_RESOURCE_WORKLOAD = "enable_resource_workload"
    ENABLE_FIRST_LAST_ACTIVITY_INDEX = "enable_first_last_activity_index"
    ENABLE_MAX_CONCURRENT_EVENTS = "enable_max_concurrent_events"
    ENABLE_MAX_CONCURRENT_EVENTS_PER_ACTIVITY = (
        "enable_max_concurrent_events_per_activity"
    )
    CASE_ATTRIBUTE_PREFIX = constants.CASE_ATTRIBUTE_PREFIX


def max_concurrent_events(
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Tuple[Any, List[str]]:
    """
    Counts for every trace the maximum number of events (of any activity) that happen concurrently
    (e.g., their time intervals [st1, ct1] and [st2, ct2] have non-empty intersection).

    Parameters
    -----------------
    log
        Event log
    parameters
        Parameters of the algorithm

    Returns
    ----------------
    data
        Numeric value of the features
    feature_names
        Names of the features
    """
    if parameters is None:
        parameters = {}

    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )

    data = []
    feature_names = ["@@max_concurrent_activities_general"]

    for trace in log:
        max_conc = 0
        i = 0
        while i < len(trace) - 1:
            conc = 0
            ct = trace[i][timestamp_key].timestamp()
            j = i + 1
            while j < len(trace):
                st = trace[j][start_timestamp_key].timestamp()
                if st > ct:
                    break
                conc = conc + 1
                j = j + 1
            if conc > max_conc:
                max_conc = conc
            i = i + 1
        data.append([float(max_conc)])

    return data, feature_names


def max_concurrent_events_per_activity(
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Tuple[Any, List[str]]:
    """
    Counts for every trace and every activity the maximum number of events of the given activity that happen concurrently
    (e.g., their time intervals [st1, ct1] and [st2, ct2] have non-empty intersection).

    Parameters
    -----------------
    log
        Event log
    parameters
        Parameters of the algorithm

    Returns
    ----------------
    data
        Numeric value of the features
    feature_names
        Names of the features
    """
    if parameters is None:
        parameters = {}

    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )

    activities = list(set(y[activity_key] for x in log for y in x))

    data = []
    feature_names = [
        "@@max_concurrent_activities_like_" + x for x in activities
    ]

    for trace in log:
        max_conc_act = {act: 0 for act in activities}
        i = 0
        while i < len(trace) - 1:
            conc = 0
            act = trace[i][activity_key]
            ct = trace[i][timestamp_key].timestamp()
            j = i + 1
            while j < len(trace):
                st = trace[j][start_timestamp_key].timestamp()
                actj = trace[j][activity_key]
                if st > ct:
                    break
                if act == actj:
                    conc = conc + 1
                j = j + 1
            i = i + 1
            max_conc_act[act] = max(max_conc_act[act], conc)
        arr = []
        for act in activities:
            arr.append(float(max_conc_act[act]))
        data.append(arr)

    return data, feature_names


def resource_workload(
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Tuple[Any, List[str]]:
    """
    Calculates for each case, and for each resource of the log, the workload of the resource during
    the lead time of a case. Defaults if a resource is not contained in a case.

    Parameters
    -----------------
    log
        Event log
    parameters
        Parameters of the algorithm

    Returns
    ----------------
    data
        Numeric value of the features
    feature_names
        Names of the features
    """
    if parameters is None:
        parameters = {}

    from pm4py.util.intervaltree import IntervalTree, Interval

    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    resource_key = exec_utils.get_param_value(
        Parameters.RESOURCE_KEY, parameters, xes_constants.DEFAULT_RESOURCE_KEY
    )
    epsilon = exec_utils.get_param_value(
        Parameters.EPSILON, parameters, 0.000001
    )
    default_not_present = exec_utils.get_param_value(
        Parameters.DEFAULT_NOT_PRESENT, parameters, 0
    )

    tree_dict = {}
    for case in log:
        if case:
            resources = set(x[resource_key] for x in case if resource_key in x)
            st = case[0][start_timestamp_key].timestamp() - epsilon
            ct = case[-1][timestamp_key].timestamp() + epsilon
            for res in resources:
                if res not in tree_dict:
                    tree_dict[res] = IntervalTree()
                tree_dict[res].add(Interval(st, ct))

    resources_list = sorted(list(tree_dict))

    data = []
    feature_names = ["resource_workload@@" + r for r in resources_list]

    for case in log:
        data.append([])
        resources = set(x[resource_key] for x in case if resource_key in x)
        st = case[0][start_timestamp_key].timestamp() - epsilon
        ct = case[-1][timestamp_key].timestamp() + epsilon
        for res in resources_list:
            if res in resources:
                data[-1].append(float(len(tree_dict[res][st:ct])))
            else:
                data[-1].append(float(default_not_present))

    return data, feature_names


def work_in_progress(
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Tuple[Any, List[str]]:
    """
    Calculates for each case, and for each resource of the log, the number of cases which are open during
    the lead time of the case.

    Parameters
    -----------------
    log
        Event log
    parameters
        Parameters of the algorithm

    Returns
    ----------------
    data
        Numeric value of the features
    feature_names
        Names of the features
    """
    if parameters is None:
        parameters = {}

    from pm4py.util.intervaltree import IntervalTree, Interval

    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    epsilon = exec_utils.get_param_value(
        Parameters.EPSILON, parameters, 0.000001
    )
    default_not_present = exec_utils.get_param_value(
        Parameters.DEFAULT_NOT_PRESENT, parameters, 0
    )

    tree = IntervalTree()
    for case in log:
        if case:
            st = case[0][start_timestamp_key].timestamp() - epsilon
            ct = case[-1][timestamp_key].timestamp() + epsilon
            tree.add(Interval(st, ct))

    data = []
    feature_names = ["@@work_in_progress"]

    for case in log:
        if case:
            st = case[0][start_timestamp_key].timestamp() - epsilon
            ct = case[-1][timestamp_key].timestamp() + epsilon
            data.append([float(len(tree[st:ct]))])
        else:
            data.append([float(default_not_present)])

    return data, feature_names


def indirect_paths_times_last_occ(
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Tuple[Any, List[str]]:
    """
    Calculates for each case, and for each indirect path of the case, the difference between the start timestamp
    of the later event and the completion timestamp of the first event. Defaults if a path is not present in a case.

    Parameters
    -----------------
    log
        Event log
    parameters
        Parameters of the algorithm

    Returns
    ----------------
    data
        Numeric value of the features
    feature_names
        Names of the features
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    default_not_present = exec_utils.get_param_value(
        Parameters.DEFAULT_NOT_PRESENT, parameters, 0
    )

    all_paths = set()
    for trace in log:
        for i in range(len(trace) - 1):
            for j in range(i + 2, len(trace)):
                all_paths.add((trace[i][activity_key], trace[j][activity_key]))
    all_paths = sorted(list(all_paths))

    data = []
    feature_names = []
    for p in all_paths:
        feature_names.append(
            "indirectPathPerformanceLastOcc@@" + p[0] + "##" + p[1]
        )

    for trace in log:
        data.append([])
        trace_paths_perf = {}
        for i in range(len(trace) - 1):
            for j in range(i + 2, len(trace)):
                p = (trace[i][activity_key], trace[j][activity_key])
                tc = trace[i][timestamp_key].timestamp()
                ts = trace[j][start_timestamp_key].timestamp()
                if ts > tc:
                    trace_paths_perf[p] = ts - tc
        for p in all_paths:
            if p in trace_paths_perf:
                data[-1].append(float(trace_paths_perf[p]))
            else:
                data[-1].append(float(default_not_present))

    return data, feature_names


def direct_paths_times_last_occ(
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Tuple[Any, List[str]]:
    """
    Calculates for each case, and for each direct path of the case, the difference between the start timestamp
    of the later event and the completion timestamp of the first event. Defaults if a path is not present in a case.

    Parameters
    -----------------
    log
        Event log
    parameters
        Parameters of the algorithm

    Returns
    ----------------
    data
        Numeric value of the features
    feature_names
        Names of the features
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    default_not_present = exec_utils.get_param_value(
        Parameters.DEFAULT_NOT_PRESENT, parameters, 0
    )

    all_paths = set()
    for trace in log:
        for i in range(len(trace) - 1):
            all_paths.add((trace[i][activity_key], trace[i + 1][activity_key]))
    all_paths = sorted(list(all_paths))

    data = []
    feature_names = []
    for p in all_paths:
        feature_names.append(
            "directPathPerformanceLastOcc@@" + p[0] + "##" + p[1]
        )

    for trace in log:
        data.append([])
        trace_paths_perf = {}
        for i in range(len(trace) - 1):
            p = (trace[i][activity_key], trace[i + 1][activity_key])
            tc = trace[i][timestamp_key].timestamp()
            ts = trace[i + 1][start_timestamp_key].timestamp()
            if ts > tc:
                trace_paths_perf[p] = ts - tc
        for p in all_paths:
            if p in trace_paths_perf:
                data[-1].append(float(trace_paths_perf[p]))
            else:
                data[-1].append(float(default_not_present))

    return data, feature_names


def times_from_first_occurrence_activity_case(
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Tuple[Any, List[str]]:
    """
    Calculates for each case, and for each activity, the times from the start to the case, and to the end of the case,
    from the first occurrence of the activity in the case.

    Parameters
    -----------------
    log
        Event log
    parameters
        Parameters of the algorithm

    Returns
    ----------------
    data
        Numeric value of the features
    feature_names
        Names of the features
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    default_not_present = exec_utils.get_param_value(
        Parameters.DEFAULT_NOT_PRESENT, parameters, 0
    )

    activities_log = set()
    for trace in log:
        for event in trace:
            activities_log.add(event[activity_key])
    activities_log = sorted(list(activities_log))

    data = []
    feature_names = []
    for act in activities_log:
        feature_names.append("startToFirstOcc@@" + act)
        feature_names.append("firstOccToEnd@@" + act)

    for trace in log:
        data.append([])
        activities_occ = {}
        for i in range(len(trace)):
            if not trace[i][activity_key] in activities_occ:
                activities_occ[trace[i][activity_key]] = i
        for act in activities_log:
            if act in activities_occ:
                ev = trace[activities_occ[act]]
                this_ev_st = ev[start_timestamp_key].timestamp()
                this_ev_ct = ev[timestamp_key].timestamp()
                start_ev_ct = trace[0][timestamp_key].timestamp()
                end_ev_st = trace[-1][start_timestamp_key].timestamp()
                data[-1].append(float(this_ev_st - start_ev_ct))
                data[-1].append(float(end_ev_st - this_ev_ct))
            else:
                data[-1].append(float(default_not_present))
                data[-1].append(float(default_not_present))

    return data, feature_names


def times_from_last_occurrence_activity_case(
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Tuple[Any, List[str]]:
    """
    Calculates for each case, and for each activity, the times from the start to the case, and to the end of the case,
    from the last occurrence of the activity in the case.

    Parameters
    -----------------
    log
        Event log
    parameters
        Parameters of the algorithm

    Returns
    ----------------
    data
        Numeric value of the features
    feature_names
        Names of the features
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    default_not_present = exec_utils.get_param_value(
        Parameters.DEFAULT_NOT_PRESENT, parameters, 0
    )

    activities_log = set()
    for trace in log:
        for event in trace:
            activities_log.add(event[activity_key])
    activities_log = sorted(list(activities_log))

    data = []
    feature_names = []
    for act in activities_log:
        feature_names.append("startToLastOcc@@" + act)
        feature_names.append("lastOccToEnd@@" + act)

    for trace in log:
        data.append([])
        activities_occ = {}
        for i in range(len(trace)):
            activities_occ[trace[i][activity_key]] = i
        for act in activities_log:
            if act in activities_occ:
                ev = trace[activities_occ[act]]
                this_ev_st = ev[start_timestamp_key].timestamp()
                this_ev_ct = ev[timestamp_key].timestamp()
                start_ev_ct = trace[0][timestamp_key].timestamp()
                end_ev_st = trace[-1][start_timestamp_key].timestamp()
                data[-1].append(float(this_ev_st - start_ev_ct))
                data[-1].append(float(end_ev_st - this_ev_ct))
            else:
                data[-1].append(float(default_not_present))
                data[-1].append(float(default_not_present))

    return data, feature_names


def first_last_activity_index_trace(
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Tuple[Any, List[str]]:
    """
    Consider as features the first and the last index of an activity inside a case

    Parameters
    ------------------
    log
        Event log
    parameters
        Parameters, including:
        - Parameters.ACTIVITY_KEY => the attribute to use as activity
        - Parameters.DEFAULT_NOT_PRESENT => the replacement value for activities that are not present for the specific case

    Returns
    -----------------
    data
        Numeric value of the features
    feature_names
        Names of the features
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    default_not_present = exec_utils.get_param_value(
        Parameters.DEFAULT_NOT_PRESENT, parameters, -1
    )

    activities_log = set()
    for trace in log:
        for event in trace:
            activities_log.add(event[activity_key])
    activities_log = sorted(list(activities_log))

    data = []
    feature_names = []
    for act in activities_log:
        feature_names.append("firstIndexAct@@" + act)
        feature_names.append("lastIndexAct@@" + act)
    for trace in log:
        data.append([])

        first_occ = {}
        last_occ = {}
        for index, event in enumerate(trace):
            act = event[activity_key]
            last_occ[act] = index
            if act not in first_occ:
                first_occ[act] = index
        for act in activities_log:
            if act not in first_occ:
                data[-1].append(float(default_not_present))
                data[-1].append(float(default_not_present))
            else:
                data[-1].append(float(first_occ[act]))
                data[-1].append(float(last_occ[act]))

    return data, feature_names


def case_duration(
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Tuple[Any, List[str]]:
    """
    Calculates for each case, the case duration (and adds it as a feature)

    Parameters
    -----------------
    log
        Event log
    parameters
        Parameters of the algorithm

    Returns
    ----------------
    data
        Numeric value of the features
    feature_names
        Names of the features
    """
    if parameters is None:
        parameters = {}

    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )

    feature_names = ["@@caseDuration"]
    data = []
    for trace in log:
        if trace:
            data.append(
                [
                    float(
                        trace[-1][timestamp_key].timestamp()
                        - trace[0][start_timestamp_key].timestamp()
                    )
                ]
            )
        else:
            data.append([0.0])

    return data, feature_names


def get_string_trace_attribute_rep(trace: Trace, trace_attribute: str) -> str:
    """
    Get a representation of the feature name associated to a string trace attribute value

    Parameters
    ------------
    trace
        Trace of the log
    trace_attribute
        Attribute of the trace to consider

    Returns
    ------------
    rep
        Representation of the feature name associated to a string trace attribute value
    """
    if trace_attribute in trace.attributes:
        return (
            "trace:"
            + str(trace_attribute)
            + "@"
            + str(trace.attributes[trace_attribute])
        )
    return "trace:" + str(trace_attribute) + "@UNDEFINED"


def get_all_string_trace_attribute_values(
    log: EventLog, trace_attribute: str
) -> List[str]:
    """
    Get all string trace attribute values representations for a log

    Parameters
    ------------
    log
        Trace log
    trace_attribute
        Attribute of the trace to consider

    Returns
    ------------
    list
        List containing for each trace a representation of the feature name associated to the attribute
    """
    values = set()
    for trace in log:
        values.add(get_string_trace_attribute_rep(trace, trace_attribute))
    return list(sorted(values))


def get_string_event_attribute_rep(event: Event, event_attribute: str) -> str:
    """
    Get a representation of the feature name associated to a string event attribute value

    Parameters
    ------------
    event
        Single event of a trace
    event_attribute
        Event attribute to consider

    Returns
    ------------
    rep
        Representation of the feature name associated to a string event attribute value
    """
    return "event:" + str(event_attribute) + "@" + str(event[event_attribute])


def get_values_event_attribute_for_trace(
    trace: Trace, event_attribute: str
) -> Set[str]:
    """
    Get all the representations for the events of a trace associated to a string event attribute values

    Parameters
    -------------
    trace
        Trace of the log
    event_attribute
        Event attribute to consider

    Returns
    -------------
    values
        All feature names present for the given attribute in the given trace
    """
    values_trace = set()
    for event in trace:
        if event_attribute in event:
            values_trace.add(
                get_string_event_attribute_rep(event, event_attribute)
            )
    if not values_trace:
        values_trace.add("event:" + str(event_attribute) + "@UNDEFINED")
    return values_trace


def get_all_string_event_attribute_values(
    log: EventLog, event_attribute: str
) -> List[str]:
    """
    Get all the representations for all the traces of the log associated to a string event attribute values

    Parameters
    ------------
    log
        Trace of the log
    event_attribute
        Event attribute to consider

    Returns
    ------------
    values
        All feature names present for the given attribute in the given log
    """
    values = set()
    for trace in log:
        values = values.union(
            get_values_event_attribute_for_trace(trace, event_attribute)
        )
    return list(sorted(values))


def get_string_event_attribute_succession_rep(
    event1: Event, event2: Event, event_attribute: str
) -> str:
    """
    Get a representation of the feature name associated to a string event attribute value

    Parameters
    ------------
    event1
        First event of the succession
    event2
        Second event of the succession
    event_attribute
        Event attribute to consider

    Returns
    ------------
    rep
        Representation of the feature name associated to a string event attribute value
    """
    return (
        "succession:"
        + str(event_attribute)
        + "@"
        + str(event1[event_attribute])
        + "#"
        + str(event2[event_attribute])
    )


def get_values_event_attribute_succession_for_trace(
    trace: Trace, event_attribute: str
) -> Set[str]:
    """
    Get all the representations for the events of a trace associated to a string event attribute succession values

    Parameters
    -------------
    trace
        Trace of the log
    event_attribute
        Event attribute to consider

    Returns
    -------------
    values
        All feature names present for the given attribute succession in the given trace
    """
    values_trace = set()
    for i in range(len(trace) - 1):
        event1 = trace[i]
        event2 = trace[i + 1]
        if event_attribute in event1 and event_attribute in event2:
            values_trace.add(
                get_string_event_attribute_succession_rep(
                    event1, event2, event_attribute
                )
            )
    if not values_trace:
        values_trace.add("succession:" + str(event_attribute) + "@UNDEFINED")
    return values_trace


def get_all_string_event_succession_attribute_values(
    log: EventLog, event_attribute: str
) -> List[str]:
    """
    Get all the representations for all the traces of the log associated to a string event attribute succession values

    Parameters
    ------------
    log
        Trace of the log
    event_attribute
        Event attribute to consider

    Returns
    ------------
    values
        All feature names present for the given attribute succession in the given log
    """
    values = set()
    for trace in log:
        values = values.union(
            get_values_event_attribute_succession_for_trace(
                trace, event_attribute
            )
        )
    return list(sorted(values))


def get_numeric_trace_attribute_rep(trace_attribute: str) -> str:
    """
    Get the feature name associated to a numeric trace attribute

    Parameters
    ------------
    trace_attribute
        Name of the trace attribute

    Returns
    ------------
    feature_name
        Name of the feature
    """
    return "trace:" + trace_attribute


def get_numeric_trace_attribute_value(
    trace: Trace, trace_attribute: str
) -> Union[int, float]:
    """
    Get the value of a numeric trace attribute from a given trace

    Parameters
    ------------
    trace
        Trace of the log

    Returns
    ------------
    value
        Value of the numeric trace attribute for the given trace
    """
    if trace_attribute in trace.attributes:
        return float(trace.attributes[trace_attribute])
    raise Exception(
        "at least a trace without trace attribute: " + trace_attribute
    )


def get_numeric_event_attribute_rep(event_attribute: str) -> str:
    """
    Get the feature name associated to a numeric event attribute

    Parameters
    ------------
    event_attribute
        Name of the event attribute

    Returns
    -------------
    feature_name
        Name of the feature
    """
    return "event:" + event_attribute


def get_numeric_event_attribute_value(
    event: Event, event_attribute: str
) -> Union[int, float]:
    """
    Get the value of a numeric event attribute from a given event

    Parameters
    -------------
    event
        Event

    Returns
    -------------
    value
        Value of the numeric event attribute for the given event
    """
    if event_attribute in event:
        return float(event[event_attribute])
    return None


def get_numeric_event_attribute_value_trace(
    trace: Trace, event_attribute: str
) -> Union[int, float]:
    """
    Get the value of the last occurrence of a numeric event attribute given a trace

    Parameters
    -------------
    trace
        Trace of the log

    Returns
    -------------
    value
        Value of the last occurrence of a numeric trace attribute for the given trace
    """
    non_zero_values = []
    for event in trace:
        value = get_numeric_event_attribute_value(event, event_attribute)
        if value is not None:
            non_zero_values.append(value)
    if len(non_zero_values) > 0:
        return non_zero_values[-1]
    raise Exception(
        "at least a trace without any event with event attribute: "
        + event_attribute
    )


def get_default_representation_with_attribute_names(
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
    feature_names: Optional[List[str]] = None,
) -> Tuple[Any, List[str], List[str], List[str], List[str], List[str]]:
    """
    Gets the default data representation of an event log (for process tree building)
    returning also the attribute names

    Parameters
    -------------
    log
        Trace log
    parameters
        Possible parameters of the algorithm
    feature_names
        (If provided) Feature to use in the representation of the log

    Returns
    -------------
    data
        Data to provide for decision tree learning
    feature_names
        Names of the features, in order
    """
    from pm4py.statistics.attributes.log.select import (
        select_attributes_from_log_for_tree,
    )

    if parameters is None:
        parameters = {}

    enable_activity_def_representation = exec_utils.get_param_value(
        Parameters.ENABLE_ACTIVITY_DEF_REPRESENTATION, parameters, False
    )
    enable_succ_def_representation = exec_utils.get_param_value(
        Parameters.ENABLE_SUCC_DEF_REPRESENTATION, parameters, False
    )

    activity_key = (
        parameters[constants.PARAMETER_CONSTANT_ACTIVITY_KEY]
        if constants.PARAMETER_CONSTANT_ACTIVITY_KEY in parameters
        else xes.DEFAULT_NAME_KEY
    )
    blacklist = parameters["blacklist"] if "blacklist" in parameters else []

    str_tr_attr, str_ev_attr, num_tr_attr, num_ev_attr = (
        select_attributes_from_log_for_tree(log)
    )
    str_evsucc_attr = None

    if enable_succ_def_representation:
        str_evsucc_attr = [activity_key]
    if enable_activity_def_representation and activity_key not in str_ev_attr:
        str_ev_attr.append(activity_key)

    str_tr_attr = [x for x in str_tr_attr if x not in blacklist]
    str_ev_attr = [x for x in str_ev_attr if x not in blacklist]
    num_tr_attr = [x for x in num_tr_attr if x not in blacklist]
    num_ev_attr = [x for x in num_ev_attr if x not in blacklist]
    if str_evsucc_attr is not None:
        str_evsucc_attr = [x for x in str_evsucc_attr if x not in blacklist]

    data, feature_names = get_representation(
        log,
        str_tr_attr,
        str_ev_attr,
        num_tr_attr,
        num_ev_attr,
        str_evsucc_attr=str_evsucc_attr,
        feature_names=feature_names,
    )

    return (
        data,
        feature_names,
        str_tr_attr,
        str_ev_attr,
        num_tr_attr,
        num_ev_attr,
    )


def get_default_representation(
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
    feature_names: Optional[List[str]] = None,
) -> Tuple[Any, List[str]]:
    """
    Gets the default data representation of an event log (for process tree building)

    Parameters
    -------------
    log
        Trace log
    parameters
        Possible parameters of the algorithm
    feature_names
        (If provided) Feature to use in the representation of the log

    Returns
    -------------
    data
        Data to provide for decision tree learning
    feature_names
        Names of the features, in order
    """
    from pm4py.statistics.attributes.log.select import (
        select_attributes_from_log_for_tree,
    )

    if parameters is None:
        parameters = {}

    enable_activity_def_representation = exec_utils.get_param_value(
        Parameters.ENABLE_ACTIVITY_DEF_REPRESENTATION, parameters, True
    )
    enable_succ_def_representation = exec_utils.get_param_value(
        Parameters.ENABLE_SUCC_DEF_REPRESENTATION, parameters, True
    )

    activity_key = (
        parameters[constants.PARAMETER_CONSTANT_ACTIVITY_KEY]
        if constants.PARAMETER_CONSTANT_ACTIVITY_KEY in parameters
        else xes.DEFAULT_NAME_KEY
    )
    blacklist = parameters["blacklist"] if "blacklist" in parameters else []

    str_tr_attr, str_ev_attr, num_tr_attr, num_ev_attr = (
        select_attributes_from_log_for_tree(log)
    )
    str_evsucc_attr = None

    if enable_succ_def_representation:
        str_evsucc_attr = [activity_key]
    if enable_activity_def_representation and activity_key not in str_ev_attr:
        str_ev_attr.append(activity_key)

    str_tr_attr = [x for x in str_tr_attr if x not in blacklist]
    str_ev_attr = [x for x in str_ev_attr if x not in blacklist]
    num_tr_attr = [x for x in num_tr_attr if x not in blacklist]
    num_ev_attr = [x for x in num_ev_attr if x not in blacklist]
    if str_evsucc_attr is not None:
        str_evsucc_attr = [x for x in str_evsucc_attr if x not in blacklist]

    return get_representation(
        log,
        str_tr_attr,
        str_ev_attr,
        num_tr_attr,
        num_ev_attr,
        str_evsucc_attr=str_evsucc_attr,
        feature_names=feature_names,
    )


def get_representation(
    log: EventLog,
    str_tr_attr: List[str],
    str_ev_attr: List[str],
    num_tr_attr: List[str],
    num_ev_attr: List[str],
    str_evsucc_attr: Optional[List[str]] = None,
    feature_names: Optional[List[str]] = None,
) -> Tuple[Any, List[str]]:
    """
    Get a representation of the event log that is suited for the data part of the decision tree learning

    NOTE: this function only encodes the last value seen for each attribute

    Parameters
    -------------
    log
        Trace log
    str_tr_attr
        List of string trace attributes to consider in data vector creation
    str_ev_attr
        List of string event attributes to consider in data vector creation
    num_tr_attr
        List of numeric trace attributes to consider in data vector creation
    num_ev_attr
        List of numeric event attributes to consider in data vector creation
    str_evsucc_attr
        List of attributes succession of values to consider in data vector creation
    feature_names
        (If provided) Feature to use in the representation of the log

    Returns
    -------------
    data
        Data to provide for decision tree learning
    feature_names
        Names of the features, in order
    """
    data = []
    dictionary = {}
    count = 0
    if feature_names is None:
        feature_names = []
        for trace_attribute in str_tr_attr:
            values = get_all_string_trace_attribute_values(
                log, trace_attribute
            )
            for value in values:
                dictionary[value] = count
                feature_names.append(value)
                count = count + 1
        for event_attribute in str_ev_attr:
            values = get_all_string_event_attribute_values(
                log, event_attribute
            )
            for value in values:
                dictionary[value] = count
                feature_names.append(value)
                count = count + 1
        for trace_attribute in num_tr_attr:
            dictionary[get_numeric_trace_attribute_rep(trace_attribute)] = (
                count
            )
            feature_names.append(
                get_numeric_trace_attribute_rep(trace_attribute)
            )
            count = count + 1
        for event_attribute in num_ev_attr:
            dictionary[get_numeric_event_attribute_rep(event_attribute)] = (
                count
            )
            feature_names.append(
                get_numeric_event_attribute_rep(event_attribute)
            )
            count = count + 1
        if str_evsucc_attr:
            for event_attribute in str_evsucc_attr:
                values = get_all_string_event_succession_attribute_values(
                    log, event_attribute
                )
                for value in values:
                    dictionary[value] = count
                    feature_names.append(value)
                    count = count + 1
    else:
        count = len(feature_names)
        for index, value in enumerate(feature_names):
            dictionary[value] = index

    for trace in log:
        trace_rep = [0.0] * count
        for trace_attribute in str_tr_attr:
            trace_attr_rep = get_string_trace_attribute_rep(
                trace, trace_attribute
            )
            if trace_attr_rep in dictionary:
                trace_rep[dictionary[trace_attr_rep]] = 1.0
        for event_attribute in str_ev_attr:
            values = get_values_event_attribute_for_trace(
                trace, event_attribute
            )
            for value in values:
                if value in dictionary:
                    trace_rep[dictionary[value]] = 1.0
        for trace_attribute in num_tr_attr:
            this_value = get_numeric_trace_attribute_rep(trace_attribute)
            if this_value in dictionary:
                trace_rep[dictionary[this_value]] = (
                    get_numeric_trace_attribute_value(trace, trace_attribute)
                )
        for event_attribute in num_ev_attr:
            this_value = get_numeric_event_attribute_rep(event_attribute)
            if this_value in dictionary:
                trace_rep[dictionary[this_value]] = (
                    get_numeric_event_attribute_value_trace(
                        trace, event_attribute
                    )
                )
        if str_evsucc_attr:
            for event_attribute in str_evsucc_attr:
                values = get_values_event_attribute_succession_for_trace(
                    trace, event_attribute
                )
                for value in values:
                    if value in dictionary:
                        trace_rep[dictionary[value]] = 1.0
        data.append(trace_rep)
    # data = np.asarray(data)
    return data, feature_names


def apply(
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Tuple[Any, List[str]]:
    """
    Extract the features from an event log (a vector for each trace)

    Parameters
    -----------------
    log
        Log
    parameters
        Parameters of the algorithm, including:
        - STR_TRACE_ATTRIBUTES => string trace attributes to consider in the features extraction
        - STR_EVENT_ATTRIBUTES => string event attributes to consider in the features extraction
        - NUM_TRACE_ATTRIBUTES => numeric trace attributes to consider in the features extraction
        - NUM_EVENT_ATTRIBUTES => numeric event attributes to consider in the features extraction
        - STR_EVSUCC_ATTRIBUTES => succession of event attributes to consider in the features extraction
        - FEATURE_NAMES => features to consider (in the given order)
        - ENABLE_ALL_EXTRA_FEATURES => enables all the extra features
        - ENABLE_CASE_DURATION => enables the case duration as additional feature
        - ENABLE_TIMES_FROM_FIRST_OCCURRENCE => enables the addition of the times from start of the case, to the end
        of the case, from the first occurrence of an activity of a case
        - ADD_CASE_IDENTIFIER_COLUMN => adds the case identifier (string) as column of the feature table (default: False)
        - ENABLE_TIMES_FROM_LAST_OCCURRENCE => enables the addition of the times from start of the case, to the end
        of the case, from the last occurrence of an activity of a case
        - ENABLE_DIRECT_PATHS_TIMES_LAST_OCC => add the duration of the last occurrence of a directed (i, i+1) path
        in the case as feature
        - ENABLE_INDIRECT_PATHS_TIMES_LAST_OCC => add the duration of the last occurrence of an indirect (i, j) path
        in the case as feature
        - ENABLE_WORK_IN_PROGRESS => enables the work in progress (number of concurrent cases) as a feature
        - ENABLE_RESOURCE_WORKLOAD => enables the resource workload as a feature
        - ENABLE_FIRST_LAST_ACTIVITY_INDEX => enables the insertion of the indexes of the activities as features
        - ENABLE_MAX_CONCURRENT_EVENTS => enables the count of the number of concurrent events inside a case
        - ENABLE_MAX_CONCURRENT_EVENTS_PER_ACTIVITY => enables the count of the number of concurrent events per activity

    Returns
    -------------
    data
        Data to provide for decision tree learning
    feature_names
        Names of the features, in order
    """
    if parameters is None:
        parameters = {}

    str_tr_attr = exec_utils.get_param_value(
        Parameters.STR_TRACE_ATTRIBUTES, parameters, None
    )
    num_tr_attr = exec_utils.get_param_value(
        Parameters.NUM_TRACE_ATTRIBUTES, parameters, None
    )
    str_ev_attr = exec_utils.get_param_value(
        Parameters.STR_EVENT_ATTRIBUTES, parameters, None
    )
    num_ev_attr = exec_utils.get_param_value(
        Parameters.NUM_EVENT_ATTRIBUTES, parameters, None
    )
    str_evsucc_attr = exec_utils.get_param_value(
        Parameters.STR_EVSUCC_ATTRIBUTES, parameters, None
    )
    feature_names = exec_utils.get_param_value(
        Parameters.FEATURE_NAMES, parameters, None
    )

    at_least_one_provided = (
        (str_tr_attr is not None)
        or (num_tr_attr is not None)
        or (str_ev_attr is not None)
        or (num_ev_attr is not None)
    )

    if str_tr_attr is None:
        str_tr_attr = []

    if num_tr_attr is None:
        num_tr_attr = []

    if str_ev_attr is None:
        str_ev_attr = []

    if num_ev_attr is None:
        num_ev_attr = []

    if pandas_utils.check_is_pandas_dataframe(log):
        case_attribute_prefix = exec_utils.get_param_value(
            Parameters.CASE_ATTRIBUTE_PREFIX, parameters, "case:"
        )

        if str_tr_attr or num_tr_attr or str_ev_attr or num_ev_attr:
            columns = list(
                set([case_attribute_prefix + x for x in str_tr_attr])
                .union(set([case_attribute_prefix + x for x in num_tr_attr]))
                .union(set(str_ev_attr))
                .union(set(num_ev_attr))
            )
            if is_polars_lazyframe(log):
                from pm4py.objects.log.util import pl_lazy_fea_utils
                fea_df = pl_lazy_fea_utils.get_features_df(log, columns, parameters=parameters)
            else:
                fea_df = dataframe_utils.get_features_df(
                    log, columns, parameters=parameters
                )
            if is_polars_lazyframe(fea_df):
                feature_names = list(fea_df.collect_schema().names())
            else:
                feature_names = list(fea_df.columns)
        else:
            if is_polars_lazyframe(log):
                from pm4py.objects.log.util import pl_lazy_fea_utils
                fea_df = pl_lazy_fea_utils.automatic_feature_extraction_df(log, parameters=parameters)
            else:
                fea_df = dataframe_utils.automatic_feature_extraction_df(
                    log, parameters=parameters
                )
            if is_polars_lazyframe(fea_df):
                feature_names = list(fea_df.collect_schema().names())
            else:
                feature_names = list(fea_df.columns)
        return fea_df, feature_names
    else:
        enable_all = exec_utils.get_param_value(
            Parameters.ENABLE_ALL_EXTRA_FEATURES, parameters, False
        )
        case_id_key = exec_utils.get_param_value(
            Parameters.CASE_ID_KEY,
            parameters,
            xes_constants.DEFAULT_TRACEID_KEY,
        )
        add_case_identifier_column = exec_utils.get_param_value(
            Parameters.ADD_CASE_IDENTIFIER_COLUMN, parameters, False
        )
        enable_case_duration = exec_utils.get_param_value(
            Parameters.ENABLE_CASE_DURATION, parameters, enable_all
        )
        enable_times_from_first_occ = exec_utils.get_param_value(
            Parameters.ENABLE_TIMES_FROM_FIRST_OCCURRENCE,
            parameters,
            enable_all,
        )
        enable_times_from_last_occ = exec_utils.get_param_value(
            Parameters.ENABLE_TIMES_FROM_LAST_OCCURRENCE,
            parameters,
            enable_all,
        )
        enable_direct_paths_times_last_occ = exec_utils.get_param_value(
            Parameters.ENABLE_DIRECT_PATHS_TIMES_LAST_OCC,
            parameters,
            enable_all,
        )
        enable_indirect_paths_times_last_occ = exec_utils.get_param_value(
            Parameters.ENABLE_INDIRECT_PATHS_TIMES_LAST_OCC,
            parameters,
            enable_all,
        )
        enable_work_in_progress = exec_utils.get_param_value(
            Parameters.ENABLE_WORK_IN_PROGRESS, parameters, enable_all
        )
        enable_resource_workload = exec_utils.get_param_value(
            Parameters.ENABLE_RESOURCE_WORKLOAD, parameters, enable_all
        )
        enable_first_last_activity_index = exec_utils.get_param_value(
            Parameters.ENABLE_FIRST_LAST_ACTIVITY_INDEX, parameters, enable_all
        )
        enable_max_concurrent_events = exec_utils.get_param_value(
            Parameters.ENABLE_MAX_CONCURRENT_EVENTS, parameters, enable_all
        )
        enable_max_concurrent_events_per_activity = exec_utils.get_param_value(
            Parameters.ENABLE_MAX_CONCURRENT_EVENTS_PER_ACTIVITY,
            parameters,
            enable_all,
        )

        log = converter.apply(
            log, variant=converter.Variants.TO_EVENT_LOG, parameters=parameters
        )
        if at_least_one_provided:
            datas, features_namess = get_representation(
                log,
                str_tr_attr,
                str_ev_attr,
                num_tr_attr,
                num_ev_attr,
                str_evsucc_attr=str_evsucc_attr,
                feature_names=feature_names,
            )
        else:
            datas, features_namess = get_default_representation(
                log, parameters=parameters
            )

        if add_case_identifier_column:
            for i in range(len(datas)):
                datas[i] = [log[i].attributes[case_id_key]] + datas[i]
            features_namess = ["@@case_id_column"] + features_namess

        # add additional features

        if enable_case_duration:
            data, features_names = case_duration(log, parameters=parameters)
            for i in range(len(datas)):
                datas[i] = datas[i] + data[i]
            features_namess = features_namess + features_names

        if enable_times_from_first_occ:
            data, features_names = times_from_first_occurrence_activity_case(
                log, parameters=parameters
            )
            for i in range(len(datas)):
                datas[i] = datas[i] + data[i]
            features_namess = features_namess + features_names

        if enable_times_from_last_occ:
            data, features_names = times_from_last_occurrence_activity_case(
                log, parameters=parameters
            )
            for i in range(len(datas)):
                datas[i] = datas[i] + data[i]
            features_namess = features_namess + features_names

        if enable_direct_paths_times_last_occ:
            data, features_names = direct_paths_times_last_occ(
                log, parameters=parameters
            )
            for i in range(len(datas)):
                datas[i] = datas[i] + data[i]
            features_namess = features_namess + features_names

        if enable_indirect_paths_times_last_occ:
            data, features_names = indirect_paths_times_last_occ(
                log, parameters=parameters
            )
            for i in range(len(datas)):
                datas[i] = datas[i] + data[i]
            features_namess = features_namess + features_names

        if enable_work_in_progress:
            data, features_names = work_in_progress(log, parameters=parameters)
            for i in range(len(datas)):
                datas[i] = datas[i] + data[i]
            features_namess = features_namess + features_names

        if enable_resource_workload:
            data, features_names = resource_workload(
                log, parameters=parameters
            )
            for i in range(len(datas)):
                datas[i] = datas[i] + data[i]
            features_namess = features_namess + features_names

        if enable_first_last_activity_index:
            data, features_names = first_last_activity_index_trace(
                log, parameters=parameters
            )
            for i in range(len(datas)):
                datas[i] = datas[i] + data[i]
            features_namess = features_namess + features_names

        if enable_max_concurrent_events:
            data, features_names = max_concurrent_events(
                log, parameters=parameters
            )
            for i in range(len(datas)):
                datas[i] = datas[i] + data[i]
            features_namess = features_namess + features_names

        if enable_max_concurrent_events_per_activity:
            data, features_names = max_concurrent_events_per_activity(
                log, parameters=parameters
            )
            for i in range(len(datas)):
                datas[i] = datas[i] + data[i]
            features_namess = features_namess + features_names

        return datas, features_namess



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import random
import time

from typing import Optional, Set
from copy import copy, deepcopy

from pm4py.objects.log.obj import Trace, Event
from pm4py.objects.petri_net import semantics, properties
from pm4py.objects.petri_net.obj import (
    PetriNet,
    Marking,
    ResetNet,
    InhibitorNet,
)
from pm4py.objects.petri_net.saw_net.obj import StochasticArcWeightNet
from pm4py.util import xes_constants as xes_util


def is_sub_marking(sub_marking: Marking, marking: Marking) -> bool:
    for p in sub_marking:
        if p not in marking:
            return False
        elif marking[p] > sub_marking[p]:
            return False
    return True


def place_set_as_marking(places) -> Marking:
    m = Marking()
    for p in places:
        m[p] = 1
    return m


def get_arc_type(elem):
    if properties.ARCTYPE in elem.properties:
        return elem.properties[properties.ARCTYPE]
    return None


def pre_set(elem, arc_type=None) -> Set:
    pre = set()
    for a in elem.in_arcs:
        if get_arc_type(a) == arc_type:
            pre.add(a.source)
    return pre


def post_set(elem, arc_type=None) -> Set:
    post = set()
    for a in elem.out_arcs:
        if get_arc_type(a) == arc_type:
            post.add(a.target)
    return post


def remove_transition(net: PetriNet, trans: PetriNet.Transition) -> PetriNet:
    """
    Remove a transition from a Petri net

    Parameters
    ----------
    net
        Petri net
    trans
        Transition to remove

    Returns
    ----------
    net
        Petri net
    """
    if trans in net.transitions:
        in_arcs = trans.in_arcs
        for arc in in_arcs:
            place = arc.source
            place.out_arcs.remove(arc)
            net.arcs.remove(arc)
        out_arcs = trans.out_arcs
        for arc in out_arcs:
            place = arc.target
            place.in_arcs.remove(arc)
            net.arcs.remove(arc)
        net.transitions.remove(trans)
    return net


def add_place(net: PetriNet, name=None) -> PetriNet.Place:
    name = (
        name
        if name is not None
        else "p_"
        + str(len(net.places))
        + "_"
        + str(time.time())
        + str(random.randint(0, 10000))
    )
    p = PetriNet.Place(name=name)
    net.places.add(p)
    return p


def add_transition(
    net: PetriNet, name=None, label=None
) -> PetriNet.Transition:
    name = (
        name
        if name is not None
        else "t_"
        + str(len(net.transitions))
        + "_"
        + str(time.time())
        + str(random.randint(0, 10000))
    )
    t = PetriNet.Transition(name=name, label=label)
    net.transitions.add(t)
    return t


def merge(trgt: Optional[PetriNet] = None, nets=None) -> PetriNet:
    trgt = trgt if trgt is not None else PetriNet()
    nets = nets if nets is not None else list()
    for net in nets:
        trgt.transitions.update(net.transitions)
        trgt.places.update(net.places)
        trgt.arcs.update(net.arcs)
    return trgt


def remove_place(net: PetriNet, place: PetriNet.Place) -> PetriNet:
    """
    Remove a place from a Petri net

    Parameters
    -------------
    net
        Petri net
    place
        Place to remove

    Returns
    -------------
    net
        Petri net
    """
    if place in net.places:
        in_arcs = place.in_arcs
        for arc in in_arcs:
            trans = arc.source
            trans.out_arcs.remove(arc)
            net.arcs.remove(arc)
        out_arcs = place.out_arcs
        for arc in out_arcs:
            trans = arc.target
            trans.in_arcs.remove(arc)
            net.arcs.remove(arc)
        net.places.remove(place)
    return net


def add_arc_from_to(
    fr, to, net: PetriNet, weight=1, type=None
) -> PetriNet.Arc:
    """
    Adds an arc from a specific element to another element in some net. Assumes from and to are in the net!

    Parameters
    ----------
    fr: transition/place from
    to:  transition/place to
    net: net to use
    weight: weight associated to the arc

    Returns
    -------
    None
    """
    if type == properties.INHIBITOR_ARC:
        if isinstance(net, InhibitorNet):
            a = InhibitorNet.InhibitorArc(fr, to, weight)
            a.properties[properties.ARCTYPE] = type
        else:
            raise Exception(
                "trying to add an inhibitor arc on a traditional Petri net object."
            )
    elif type == properties.RESET_ARC:
        if isinstance(net, ResetNet):
            a = ResetNet.ResetArc(fr, to, weight)
            a.properties[properties.ARCTYPE] = type
        else:
            raise Exception(
                "trying to add a reset arc on a traditional Petri net object."
            )
    elif type == properties.STOCHASTIC_ARC:
        if isinstance(net, StochasticArcWeightNet):
            a = StochasticArcWeightNet.Arc(fr, to, weight)
            # a.properties[properties.ARCTYPE] = type
        else:
            raise Exception(
                "trying to add a stochastic arc on a traditional Petri net object."
            )
    else:
        a = PetriNet.Arc(fr, to, weight)
    net.arcs.add(a)
    fr.out_arcs.add(a)
    to.in_arcs.add(a)

    return a


def construct_trace_net(
    trace,
    trace_name_key=xes_util.DEFAULT_NAME_KEY,
    activity_key=xes_util.DEFAULT_NAME_KEY,
):
    """
    Creates a trace net, i.e. a trace in Petri net form.

    Parameters
    ----------
    trace: :class:`list` input trace, assumed to be a list of events
    trace_name_key: :class:`str` key of the attribute that defines the name of the trace
    activity_key: :class:`str` key of the attribute of the events that defines the activity name

    Returns
    -------
    tuple: :class:`tuple` of the net, initial marking and the final marking

    """
    net = PetriNet(
        "trace net of %s" % trace.attributes[trace_name_key]
        if trace_name_key in trace.attributes
        else " "
    )
    place_map = {0: PetriNet.Place("p_0")}
    net.places.add(place_map[0])
    for i in range(0, len(trace)):
        t = PetriNet.Transition(
            "t_" + trace[i][activity_key] + "_" + str(i),
            trace[i][activity_key],
        )
        # 16/02/2021: set the trace index as property of the transition of the
        # trace net
        t.properties[properties.TRACE_NET_TRANS_INDEX] = i
        net.transitions.add(t)
        place_map[i + 1] = PetriNet.Place("p_" + str(i + 1))
        # 16/02/2021: set the place index as property of the place of the trace
        # net
        place_map[i + 1].properties[properties.TRACE_NET_PLACE_INDEX] = i + 1
        net.places.add(place_map[i + 1])
        add_arc_from_to(place_map[i], t, net)
        add_arc_from_to(t, place_map[i + 1], net)
    return net, Marking({place_map[0]: 1}), Marking({place_map[len(trace)]: 1})


def construct_trace_net_cost_aware(
    trace,
    costs,
    trace_name_key=xes_util.DEFAULT_NAME_KEY,
    activity_key=xes_util.DEFAULT_NAME_KEY,
):
    """
    Creates a trace net, i.e. a trace in Petri net form mapping specific costs to transitions.

    Parameters
    ----------
    trace: :class:`list` input trace, assumed to be a list of events
    costs: :class:`list` list of costs, length should be equal to the length of the input trace
    trace_name_key: :class:`str` key of the attribute that defines the name of the trace
    activity_key: :class:`str` key of the attribute of the events that defines the activity name

    Returns
    -------
    tuple: :class:`tuple` of the net, initial marking, final marking and map of costs


    """
    net = PetriNet(
        "trace net of %s" % trace.attributes[trace_name_key]
        if trace_name_key in trace.attributes
        else " "
    )
    place_map = {0: PetriNet.Place("p_0")}
    net.places.add(place_map[0])
    cost_map = dict()
    for i in range(0, len(trace)):
        t = PetriNet.Transition(
            "t_" + trace[i][activity_key] + "_" + str(i),
            trace[i][activity_key],
        )
        # 16/02/2021: set the trace index as property of the transition of the
        # trace net
        t.properties[properties.TRACE_NET_TRANS_INDEX] = i
        cost_map[t] = costs[i]
        net.transitions.add(t)
        place_map[i + 1] = PetriNet.Place("p_" + str(i + 1))
        # 16/02/2021: set the place index as property of the place of the trace
        # net
        place_map[i + 1].properties[properties.TRACE_NET_PLACE_INDEX] = i + 1
        net.places.add(place_map[i + 1])
        add_arc_from_to(place_map[i], t, net)
        add_arc_from_to(t, place_map[i + 1], net)
    return (
        net,
        Marking({place_map[0]: 1}),
        Marking({place_map[len(trace)]: 1}),
        cost_map,
    )


def acyclic_net_variants(
    net, initial_marking, final_marking, activity_key=xes_util.DEFAULT_NAME_KEY
):
    """
    Given an acyclic accepting Petri net, initial and final marking extracts a set of variants (in form of traces)
    replayable on the net.
    Warning: this function is based on a marking exploration. If the accepting Petri net contains loops, the method
    will not work properly as it stops the search if a specific marking has already been encountered.

    Parameters
    ----------
    :param net: An acyclic workflow net
    :param initial_marking: The initial marking of the net.
    :param final_marking: The final marking of the net.
    :param activity_key: activity key to use

    Returns
    -------
    :return: variants: :class:`list` Set of variants - in the form of Trace objects - obtainable executing the net

    """
    active = {(initial_marking, ())}
    visited = set()
    variants = set()
    while active:
        curr_marking, curr_partial_trace = active.pop()
        curr_pair = (curr_marking, curr_partial_trace)
        enabled_transitions = semantics.enabled_transitions(net, curr_marking)
        for transition in enabled_transitions:
            if transition.label is not None:
                next_partial_trace = curr_partial_trace + (transition.label,)
            else:
                next_partial_trace = curr_partial_trace
            next_marking = semantics.execute(transition, net, curr_marking)
            next_pair = (next_marking, next_partial_trace)

            if next_marking == final_marking:
                variants.add(next_partial_trace)
            else:
                # If the next marking is not in visited, if the next
                # marking+partial trace is different from the current
                # one+partial trace
                if next_pair not in visited and curr_pair != next_pair:
                    active.add(next_pair)
        visited.add(curr_pair)
    trace_variants = []
    for variant in variants:
        trace = Trace()
        for activity_label in variant:
            trace.append(Event({activity_key: activity_label}))
        trace_variants.append(trace)
    return trace_variants


def get_transition_by_name(
    net: PetriNet, transition_name
) -> Optional[PetriNet.Transition]:
    """
    Get a transition by its name

    Parameters
    ------------
    net
        Petri net
    transition_name
        Transition name

    Returns
    ------------
    transition
        Transition object
    """
    for t in net.transitions:
        if t.name == transition_name:
            return t
    return None


def decorate_places_preset_trans(net: PetriNet):
    """
    Decorate places with information useful for the replay

    Parameters
    -------------
    net
        Petri net
    """
    for place in net.places:
        place.ass_trans = set()

    for trans in net.transitions:
        for place in trans.sub_marking:
            place.ass_trans.add(trans)


def decorate_transitions_prepostset(net: PetriNet):
    """
    Decorate transitions with sub and addition markings

    Parameters
    -------------
    net
        Petri net
    """
    from pm4py.objects.petri_net.obj import Marking

    for trans in net.transitions:
        sub_marking = Marking()
        add_marking = Marking()

        for arc in trans.in_arcs:
            sub_marking[arc.source] = arc.weight
            add_marking[arc.source] = -arc.weight
        for arc in trans.out_arcs:
            if arc.target in add_marking:
                add_marking[arc.target] = arc.weight + add_marking[arc.target]
            else:
                add_marking[arc.target] = arc.weight
        trans.sub_marking = sub_marking
        trans.add_marking = add_marking


def get_places_shortest_path(
    net,
    place_to_populate,
    current_place,
    places_shortest_path,
    actual_list,
    rec_depth,
    max_rec_depth,
):
    """
    Get shortest path between places lead by hidden transitions

    Parameters
    ----------
    net
        Petri net
    place_to_populate
        Place that we are populating the shortest map of
    current_place
        Current visited place (must explore its transitions)
    places_shortest_path
        Current dictionary
    actual_list
        Actual list of transitions to enable
    rec_depth
        Recursion depth
    max_rec_depth
        Maximum recursion depth
    """
    if rec_depth > max_rec_depth:
        return places_shortest_path
    if place_to_populate not in places_shortest_path:
        places_shortest_path[place_to_populate] = {}
    for t in current_place.out_arcs:
        if t.target.label is None:
            for p2 in t.target.out_arcs:
                if p2.target not in places_shortest_path[
                    place_to_populate
                ] or len(actual_list) + 1 < len(
                    places_shortest_path[place_to_populate][p2.target]
                ):
                    new_actual_list = copy(actual_list)
                    new_actual_list.append(t.target)
                    places_shortest_path[place_to_populate][p2.target] = copy(
                        new_actual_list
                    )
                    places_shortest_path = get_places_shortest_path(
                        net,
                        place_to_populate,
                        p2.target,
                        places_shortest_path,
                        new_actual_list,
                        rec_depth + 1,
                        max_rec_depth,
                    )
    return places_shortest_path


def get_places_shortest_path_by_hidden(net: PetriNet, max_rec_depth):
    """
    Get shortest path between places lead by hidden transitions

    Parameters
    ----------
    net
        Petri net
    max_rec_depth
        Maximum recursion depth
    """
    places_shortest_path = {}
    for p in net.places:
        places_shortest_path = get_places_shortest_path(
            net, p, p, places_shortest_path, [], 0, max_rec_depth
        )
    return places_shortest_path


def invert_spaths_dictionary(spaths):
    """
    Invert the shortest paths (between places) dictionary,
    from target-source to source-target

    Parameters
    -------------
    spaths
        Shortest paths dictionary

    Returns
    -------------
    inv_spaths
        Inverted shortest paths dictionary
    """
    inv_spaths = {}
    for target_place in spaths:
        for source_place in spaths[target_place]:
            if source_place not in inv_spaths:
                inv_spaths[source_place] = {}
            if target_place not in inv_spaths[source_place]:
                inv_spaths[source_place][target_place] = set()
            inv_spaths[source_place][target_place] = inv_spaths[source_place][
                target_place
            ].union(spaths[target_place][source_place])
    return inv_spaths


def remove_unconnected_components(net: PetriNet) -> PetriNet:
    """
    Remove unconnected components from a Petri net

    Parameters
    -----------
    net
        Petri net

    Returns
    -----------
    net
        Cleaned Petri net
    """
    changed_something = True
    while changed_something:
        changed_something = False
        places = list(net.places)
        for place in places:
            if len(place.in_arcs) == 0 and len(place.out_arcs) == 0:
                remove_place(net, place)
                changed_something = True
        transitions = list(net.transitions)
        for trans in transitions:
            if len(trans.in_arcs) == 0 or len(trans.out_arcs) == 0:
                remove_transition(net, trans)
                changed_something = True
    return net


def get_s_components_from_petri(
    net,
    im,
    fm,
    rec_depth=0,
    curr_s_comp=None,
    visited_places=None,
    list_s_components=None,
    max_rec_depth=6,
):
    """
    Gets the S-components from a Petri net

    Parameters
    -------------
    net
        Petri net
    im
        Initial marking
    fm
        Final marking
    curr_s_comp
        Current S component
    visited_places
        Visited places
    list_s_components
        List of S-components
    max_rec_depth
        Maximum recursion depth

    Returns
    --------------
    s_components
        List of S-components
    """
    if list_s_components is None:
        list_s_components = []
    if len(im) > 1 or len(fm) > 1:
        return list_s_components
    source = list(im.keys())[0]
    if curr_s_comp is None:
        curr_s_comp = [source]
    if visited_places is None:
        visited_places = []
    something_changed = True
    while something_changed and rec_depth < max_rec_depth:
        something_changed = False
        places_to_visit = sorted(
            list(set(curr_s_comp[len(visited_places):])),
            key=lambda x: len(x.out_arcs),
            reverse=True,
        )
        for place_to_visit in places_to_visit:
            visited_places.append(place_to_visit)
            target_trans = sorted(
                list(set([arc.target for arc in place_to_visit.out_arcs])),
                key=lambda x: len(x.out_arcs),
            )
            for trans in target_trans:
                visited_places_names = [x.name for x in visited_places]
                target_trans_target = list(
                    set(
                        [
                            arc.target
                            for arc in trans.out_arcs
                            if arc.target.name not in visited_places_names
                        ]
                    )
                )
                if target_trans_target:
                    something_changed = True
                    if len(target_trans_target) == 1:
                        new_place = target_trans_target[0]
                        curr_s_comp.append(new_place)
                    else:
                        for new_place in target_trans_target:
                            [new_curr_s_comp, new_visited_places] = deepcopy(
                                [curr_s_comp, visited_places]
                            )
                            new_curr_s_comp.append(new_place)
                            list_s_components = get_s_components_from_petri(
                                net,
                                im,
                                fm,
                                rec_depth=rec_depth + 1,
                                curr_s_comp=new_curr_s_comp,
                                visited_places=new_visited_places,
                                list_s_components=list_s_components,
                                max_rec_depth=max_rec_depth,
                            )

    if not set([place.name for place in curr_s_comp]) in list_s_components:
        list_s_components.append(set([place.name for place in curr_s_comp]))

    return list_s_components


def remove_arc(net: PetriNet, arc: PetriNet.Arc) -> PetriNet:
    """
    Removes an arc from a Petri net

    Parameters
    ---------------
    net
        Petri net
    arc
        Arc of the Petri net

    Returns
    -------------
    net
        Petri net
    """
    net.arcs.remove(arc)
    arc.source.out_arcs.remove(arc)
    arc.target.in_arcs.remove(arc)

    return net



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any


def apply(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None):
    """
    Discovers the number of new interactions between the related objects which appears in a given event.

    Parameters
    ---------------
    ocel
        OCEL
    parameters
        Parameters of the method

    Returns
    ----------------
    data
        Extracted feature values
    feature_names
        Feature names
    """
    if parameters is None:
        parameters = {}

    ordered_events = (
        parameters["ordered_events"]
        if "ordered_events" in parameters
        else ocel.events[ocel.event_id_column].to_numpy()
    )

    rel_objs = (
        ocel.relations.groupby(ocel.event_id_column)[ocel.object_id_column]
        .agg(list)
        .to_dict()
    )

    interactions = set()
    data = []
    feature_names = ["@@ev_new_interactions"]

    for ev in ordered_events:
        n = 0
        if ev in rel_objs:
            for o1 in rel_objs[ev]:
                for o2 in rel_objs[ev]:
                    if o1 < o2:
                        if not (o1, o2) in interactions:
                            n = n + 1
        data.append([float(n)])

    return data, feature_names



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from collections import Counter

import numpy as np

from pm4py.objects.petri_net.utils.reachability_graph import (
    construct_reachability_graph,
)
from pm4py.objects.stochastic_petri import tangible_reachability
from pm4py.objects.conversion.dfg import converter as dfg_converter
from pm4py.objects.random_variables import exponential, random_variable


def get_corr_hex(num):
    """
    Gets correspondence between a number
    and an hexadecimal string

    Parameters
    -------------
    num
        Number

    Returns
    -------------
    hex_string
        Hexadecimal string
    """
    if num < 10:
        return str(int(num))
    elif num < 11:
        return "A"
    elif num < 12:
        return "B"
    elif num < 13:
        return "C"
    elif num < 14:
        return "D"
    elif num < 15:
        return "E"
    elif num < 16:
        return "F"


def get_color_from_probabilities(prob_dictionary):
    """
    Returns colors from a dictionary of probabilities

    Parameters
    -------------
    prob_dictionary
        Dictionary of probabilities

    Returns
    -------------
    color_dictionary
        Dictionary of colors
    """
    color_dictionary = {}

    for state in prob_dictionary:
        value = prob_dictionary[state]
        whiteness = int(255.0 * (1.0 - value))
        w1 = whiteness / 16
        w2 = whiteness % 16
        c1 = get_corr_hex(w1)
        c2 = get_corr_hex(w2)
        color_dictionary[state] = "#FF" + c1 + c2 + c1 + c2

    return color_dictionary


def get_tangible_reachability_and_q_matrix_from_dfg_performance(
    dfg_performance, invisible_firing_rate=1000.0, parameters=None
):
    """
    Get the tangible reachability graph and the Q matrix from the performance DFG

    Parameters
    -------------
    dfg_performance
        Performance DFG
    invisible_firing_rate
        Firing rate for invisible transitions
    parameters
        Parameters

    Returns
    -------------
    reachab_graph
        Reachability graph
    tangible_reach_graph
        Tangible reachability graph
    stochastic_info
        Stochastic information
    q_matrix
        Q-matrix from the tangible reachability graph
    """
    if parameters is None:
        parameters = {}
    net, im, fm = dfg_converter.apply(dfg_performance, parameters=parameters)
    stochastic_map = {}
    for tr in net.transitions:
        if tr.label is None:
            rv = random_variable.RandomVariable()
            exp = exponential.Exponential()
            exp.scale = 1 / invisible_firing_rate
            rv.random_variable = exp
            stochastic_map[tr] = rv
        else:
            input_arc = list(tr.in_arcs)[0]
            output_arc = list(tr.out_arcs)[0]
            rv = random_variable.RandomVariable()
            el = (input_arc.source.name, output_arc.target.name)
            scale = 0
            if el in dfg_performance:
                scale = dfg_performance[el]
            if scale == 0:
                scale = 1 / invisible_firing_rate
            exp = exponential.Exponential()
            exp.scale = scale
            rv.random_variable = exp
            stochastic_map[tr] = rv
    tang_reach_graph = construct_reachability_graph(
        net, im, use_trans_name=True
    )
    q_matrix = get_q_matrix_from_tangible_exponential(
        tang_reach_graph, stochastic_map
    )
    return tang_reach_graph, tang_reach_graph, stochastic_map, q_matrix


def get_tangible_reachability_and_q_matrix_from_log_net(
    log, net, im, fm, parameters=None
):
    """
    Gets the tangible reachability graph from a log and an accepting Petri net

    Parameters
    ---------------
    log
        Event log
    net
        Petri net
    im
        Initial marking
    fm
        Final marking

    Returns
    ------------
    reachab_graph
        Reachability graph
    tangible_reach_graph
        Tangible reachability graph
    stochastic_info
        Stochastic information
    q_matrix
        Q-matrix from the tangible reachability graph
    """
    if parameters is None:
        parameters = {}
    reachability_graph, tangible_reachability_graph, stochastic_info = (
        tangible_reachability.get_tangible_reachability_from_log_net_im_fm(
            log, net, im, fm, parameters=parameters
        )
    )
    # gets the Q matrix assuming exponential distributions
    q_matrix = get_q_matrix_from_tangible_exponential(
        tangible_reachability_graph, stochastic_info
    )
    return (
        reachability_graph,
        tangible_reachability_graph,
        stochastic_info,
        q_matrix,
    )


def transient_analysis_from_petri_net_and_smap(
    net, im, s_map, delay, parameters=None
):
    """
    Gets the transient analysis from a Petri net, a stochastic map and a delay

    Parameters
    -------------
    log
        Event log
    delay
        Time delay
    parameters
        Parameters of the algorithm

    Returns
    -------------
    transient_result
        Transient analysis result
    """
    if parameters is None:
        parameters = {}
    # gets the reachability graph from the Petri net
    reachab_graph = construct_reachability_graph(net, im)
    states_reachable_from_start = set()
    for trans in reachab_graph.transitions:
        if str(trans.from_state) == "start1":
            states_reachable_from_start.add(trans.to_state)
    # get the tangible reachability graph from the reachability graph and the
    # stochastic map
    tang_reach_graph = (
        tangible_reachability.get_tangible_reachability_from_reachability(
            reachab_graph, s_map
        )
    )
    # gets the Q matrix assuming exponential distributions
    q_matrix = get_q_matrix_from_tangible_exponential(tang_reach_graph, s_map)
    states = sorted(list(tang_reach_graph.states), key=lambda x: x.name)
    states_vector = np.zeros((1, len(states)))

    for state in states_reachable_from_start:
        states_vector[0, states.index(state)] = 1.0 / len(
            states_reachable_from_start
        )

    probabilities = (
        transient_analysis_from_tangible_q_matrix_and_states_vector(
            tang_reach_graph, q_matrix, states_vector, delay
        )
    )

    color_dictionary = get_color_from_probabilities(probabilities)

    return tang_reach_graph, probabilities, color_dictionary


def get_q_matrix_from_tangible_exponential(
    tangible_reach_graph, stochastic_info
):
    """
    Gets Q matrix from tangible reachability graph and stochastic map where the
    distribution type has been forced to be exponential

    Parameters
    -----------
    tangible_reach_graph
        Tangible reachability graph
    stochastic_info
        Stochastic map for each transition

    Returns
    -----------
    q_matrix
        Q-matrix from the tangible reachability graph
    """
    stochastic_info_name = {}
    for s in stochastic_info:
        stochastic_info_name[s.name] = stochastic_info[s]

    states = sorted(list(tangible_reach_graph.states), key=lambda x: x.name)
    no_states = len(states)
    q_matrix = np.zeros((no_states, no_states))

    for i in range(no_states):
        sum_lambda = 0.0
        for trans in states[i].outgoing:
            target_state = trans.to_state
            target_state_index = states.index(target_state)
            if not target_state_index == i:
                sinfo = stochastic_info_name[trans.name]
                lambda_value = 1.0 / float(sinfo.random_variable.scale)
                sum_lambda = sum_lambda + lambda_value
                q_matrix[i, target_state_index] = (
                    q_matrix[i, target_state_index] + lambda_value
                )
        q_matrix[i, i] = -sum_lambda

    return q_matrix


def transient_analysis_from_tangible_q_matrix_and_single_state(
    tangible_reach_graph, q_matrix, source_state, time_diff
):
    """
    Do transient analysis from tangible reachability graph, Q matrix and a single state to start from

    Parameters
    -----------
    tangible_reach_graph
        Tangible reachability graph
    q_matrix
        Q matrix
    source_state
        Source state to consider
    time_diff
        Time interval we want to investigate

    Returns
    -----------
    transient_result
        Transient analysis result
    """
    states = sorted(list(tangible_reach_graph.states), key=lambda x: x.name)
    state_index = states.index(source_state)

    states_vector = np.zeros((1, len(states)))
    states_vector[0, state_index] = 1

    return transient_analysis_from_tangible_q_matrix_and_states_vector(
        tangible_reach_graph, q_matrix, states_vector, time_diff
    )


def transient_analysis_from_tangible_q_matrix_and_states_vector(
    tangible_reach_graph, q_matrix, states_vector, time_diff
):
    """
    Do transient analysis from tangible reachability graph, Q matrix and a vector of probability of states

    Parameters
    ------------
    tangible_reach_graph
        Tangible reachability graph
    q_matrix
        Q matrix
    states_vector
        Vector of states probabilities to start from
    time_diff
        Time interval we want to investigate

    Returns
    -----------
    transient_result
        Transient analysis result
    """
    from scipy.linalg import expm

    transient_result = Counter()
    states = sorted(list(tangible_reach_graph.states), key=lambda x: x.name)

    ht_matrix = expm(q_matrix * time_diff)

    res = np.matmul(states_vector, ht_matrix)
    # normalize to 1 the vector of probabilities
    res = res / np.sum(res)

    for i in range(len(states)):
        transient_result[states[i]] = res[0, i]

    return transient_result


def nullspace(a_matrix, atol=1e-13, rtol=0):
    """Compute an approximate basis for the nullspace of A.

    The algorithm used by this function is based on the singular value
    decomposition of `A`.

    Parameters
    ----------
    a_matrix : ndarray
        A should be at most 2-D.  A 1-D array with length k will be treated
        as a 2-D with shape (1, k)
    atol : float
        The absolute tolerance for a zero singular value.  Singular values
        smaller than `atol` are considered to be zero.
    rtol : float
        The relative tolerance.  Singular values less than rtol*smax are
        considered to be zero, where smax is the largest singular value.

    If both `atol` and `rtol` are positive, the combined tolerance is the
    maximum of the two; that is::
        tol = max(atol, rtol * smax)
    Singular values smaller than `tol` are considered to be zero.

    Returns
    ------------
    ns : ndarray
        If `A` is an array with shape (m, k), then `ns` will be an array
        with shape (k, n), where n is the estimated dimension of the
        nullspace of `A`.  The columns of `ns` are a basis for the
        nullspace; each element in numpy.dot(A, ns) will be approximately
        zero.
    """
    from numpy.linalg import svd

    a_matrix = np.atleast_2d(a_matrix)
    u, s, vh = svd(a_matrix)
    tol = max(atol, rtol * s[0])
    nnz = (s >= tol).sum()
    ns = vh[nnz:].conj().T
    return ns


def perform_steadystate(q_matrix, tangible_reach_graph):
    """
    Performs steady state analysis given the
    :param q_matrix:
    :return:
    """
    transient_result = Counter()
    states = sorted(list(tangible_reach_graph.states), key=lambda x: x.name)
    q_matrix_trans = np.matrix.transpose(q_matrix)
    if nullspace(q_matrix_trans).shape[1] > 0:
        M = np.matrix.transpose(nullspace(q_matrix_trans))
        vec = np.zeros(M.shape[1])
        for i in range(M.shape[0]):
            v = np.matmul(M[i], q_matrix_trans)
            val = np.sum(v) / np.sum(M[i])
            vec = vec + val * M[i]
        vec = vec / np.sum(vec)
        for i in range(len(states)):
            transient_result[states[i]] = vec[i]
        return vec
    return transient_result



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.objects.log.obj import EventLog, EventStream
import pandas as pd
from typing import Union, Optional, Dict, Any
from pm4py.objects.conversion.log import converter as log_converter
from enum import Enum
from pm4py.util import exec_utils, constants, xes_constants, pandas_utils


class Parameters(Enum):
    ARRIVAL_RATE = "arrival_rate"
    FINISH_RATE = "finish_rate"
    CASE_ID_COLUMN = constants.PARAMETER_CONSTANT_CASEID_KEY
    START_TIMESTAMP_COLUMN = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    TIMESTAMP_COLUMN = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    RESOURCE_COLUMN = constants.PARAMETER_CONSTANT_RESOURCE_KEY
    ACTIVITY_COLUMN = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    GROUPER_FREQ = "grouper_freq"
    SERVICE_TIME = "service_time"
    WAITING_TIME = "waiting_time"
    SOJOURN_TIME = "sojourn_time"
    DIFF_START_END = "diff_start_end"


def apply(
    log: Union[EventLog, EventStream, pd.DataFrame],
    parameters: Optional[Dict[Any, Any]] = None,
) -> pd.DataFrame:
    """
    Extracts temporal features with the provided granularity from the Pandas dataframe.

    Implements the approach described in the paper:
    Pourbafrani, Mahsa, Sebastiaan J. van Zelst, and Wil MP van der Aalst.
    "Supporting automatic system dynamics model generation for simulation
    in the context of process mining." International Conference on Business
    Information Systems. Springer, Cham, 2020.

    Parameters
    ---------------
    log
        Event log / Event stream / Pandas dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.GROUPER_FREQ => the time interval to be used for the grouping (default: "W")
        - Parameters.ARRIVAL_RATE => column of the dataframe which is going to host the arrival rate
        - Parameters.FINISH_RATE => column of the dataframe which is going to host the finishing rate
        - Parameters.SERVICE_TIME => column of the dataframe which is going to host the service time
        - Parameters.WAITING_TIME => column of the dataframe which is going to host the waiting time
        - Parameters.SOJOURN_TIME => column of the dataframe which is going to host the sojourn time
        - Parameters.CASE_ID_COLUMN => case ID column in the dataframe (default: case:concept:name)
        - Parameters.ACTIVITY_COLUMN => activity column in the dataframe (default: concept:name)
        - Parameters.TIMESTAMP_COLUMN => timestamp column in the dataframe (default: time:timestamp)
        - Parameters.RESOURCE_COLUMN => resource column in the dataframe (default: org:resource)
        - Parameters.START_TIMESTAMP_COLUMN => start timestamp column in the dataframe
          (if not provided, the timestamp column is used)

    Returns
    ----------------
    features_df
        Dataframe with temporal features, aggregated per time window
    """
    if parameters is None:
        parameters = {}

    grouper_freq = exec_utils.get_param_value(
        Parameters.GROUPER_FREQ, parameters, "W"
    )
    timestamp_column = exec_utils.get_param_value(
        Parameters.TIMESTAMP_COLUMN,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    start_timestamp_column = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_COLUMN, parameters, None
    )
    if start_timestamp_column is None:
        start_timestamp_column = timestamp_column

    case_id_column = exec_utils.get_param_value(
        Parameters.CASE_ID_COLUMN, parameters, constants.CASE_CONCEPT_NAME
    )
    diff_start_end = exec_utils.get_param_value(
        Parameters.DIFF_START_END, parameters, "@@diff_start_end"
    )
    arrival_rate = exec_utils.get_param_value(
        Parameters.ARRIVAL_RATE, parameters, "@@arrival_rate"
    )
    finish_rate = exec_utils.get_param_value(
        Parameters.FINISH_RATE, parameters, "@@finish_rate"
    )
    service_time = exec_utils.get_param_value(
        Parameters.SERVICE_TIME, parameters, "@@service_time"
    )
    waiting_time = exec_utils.get_param_value(
        Parameters.WAITING_TIME, parameters, "@@waiting_time"
    )
    sojourn_time = exec_utils.get_param_value(
        Parameters.SOJOURN_TIME, parameters, "@@sojourn_time"
    )
    resource_column = exec_utils.get_param_value(
        Parameters.RESOURCE_COLUMN,
        parameters,
        xes_constants.DEFAULT_RESOURCE_KEY,
    )
    activity_column = exec_utils.get_param_value(
        Parameters.ACTIVITY_COLUMN, parameters, xes_constants.DEFAULT_NAME_KEY
    )

    # Convert input event log to a DataFrame if necessary
    log_df = log_converter.apply(
        log,
        variant=log_converter.Variants.TO_DATA_FRAME,
        parameters=parameters,
    )

    # Insert arrival and finish rate columns
    log_df = pandas_utils.insert_case_arrival_finish_rate(
        log_df,
        case_id_column=case_id_column,
        timestamp_column=timestamp_column,
        arrival_rate_column=arrival_rate,
        finish_rate_column=finish_rate,
    )

    # Insert service, waiting, sojourn times
    log_df = pandas_utils.insert_case_service_waiting_time(
        log_df,
        case_id_column=case_id_column,
        timestamp_column=timestamp_column,
        diff_start_end_column=diff_start_end,
        service_time_column=service_time,
        sojourn_time_column=sojourn_time,
        waiting_time_column=waiting_time,
    )

    # Group the log by time windows
    grouped_log = log_df.groupby(
        pandas_utils.get_grouper(key=start_timestamp_column, freq=grouper_freq)
    )

    final_values = []

    for gkey, gval in grouped_log:
        dct = {}
        dct["timestamp"] = gkey

        # Group by case to get one row per case (first arrival_rate, etc.)
        gval_first = gval.groupby(case_id_column).first()

        # Keep existing features (if desired)
        dct["unique_resources"] = gval[resource_column].nunique()
        dct["unique_activities"] = gval[activity_column].nunique()
        dct["num_events"] = len(gval)

        dct["average_arrival_rate"] = gval_first[arrival_rate].mean()
        dct["average_finish_rate"] = gval_first[finish_rate].mean()
        dct["average_waiting_time"] = gval_first[waiting_time].mean()
        dct["average_sojourn_time"] = gval_first[sojourn_time].mean()
        dct["average_service_time"] = gval_first[service_time].mean()

        # -----------------------------------------------------------------
        # NEW FEATURES
        # -----------------------------------------------------------------
        # (1) Total number of reworked activities
        # For each case: (number_of_events) - (unique_activities)
        # Then sum across all cases in the window.
        if len(gval) > 0:
            reworks_series = gval.groupby(case_id_column)[activity_column].apply(
                lambda series: series.size - series.nunique()
            )
            dct["total_number_of_reworked_activities"] = reworks_series.sum()

            # (2) Average number of cases per resource
            # For each resource, count distinct cases, then average over resources
            cases_per_resource = gval.groupby(resource_column)[case_id_column].nunique()
            dct["avg_cases_per_resource"] = (
                cases_per_resource.mean() if not pd.isna(cases_per_resource.mean()) else 0
            )

            # (3) Average number of events per case
            # For each case, count events, then average
            events_per_case = gval.groupby(case_id_column).size()
            dct["avg_events_per_case"] = (
                events_per_case.mean() if not pd.isna(events_per_case.mean()) else 0
            )

            # (4) Number of cases
            dct["number_of_cases"] = gval[case_id_column].nunique()

            # (5) Average number of resources per case
            # For each case, count distinct resources, then average
            resources_per_case = gval.groupby(case_id_column)[resource_column].nunique()
            dct["avg_resources_per_case"] = (
                resources_per_case.mean() if not pd.isna(resources_per_case.mean()) else 0
            )
            # -----------------------------------------------------------------
        else:
            dct["total_number_of_reworked_activities"] = 0
            dct["avg_cases_per_resource"] = 0
            dct["avg_events_per_case"] = 0
            dct["avg_resources_per_case"] = 0

        final_values.append(dct)

    dataframe = pandas_utils.instantiate_dataframe(final_values)
    # Fill any remaining NaNs with 0
    dataframe = dataframe.fillna(0)

    return dataframe



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util import exec_utils
from enum import Enum
from typing import Optional, Dict, Any, Tuple


class Parameters(Enum):
    INCLUDE_HEADER = "include_header"


def get_model_description():
    description = "The temporal profile is a model describing the average and the standard deviation of the times between couples of activities eventually (not only directly) following each other in at least a process execution (so in a trace <A,B,C,D> the couples (A,B) (A,C) (A,D) (B,C) (B,D) (C,D) shall be considered). Given a positive value ZETA, a deviation occurs in a process execution when the time between two activities is lower than AVG - ZETA * STDEV or greater than AVG + ZETA * STDEV.\n"
    return description


def get_model_implementation():
    implementation = "The temporal profile is expressed as a Python dictionary associating to some couples of activities the average and the standard deviation of the times. Example: {('A', 'B'): (86400, 3600), ('B', 'C'): (3600, 3600)} indicates that the average time between A and B is 1 day, while the standard deviation is 1 hour. On the other hand, the average time between B and C is 1 hour, while the standard deviation is 1 hour."
    return implementation


def apply(
    temporal_profile: Dict[Tuple[str, str], Tuple[float, float]],
    parameters: Optional[Dict[Any, Any]] = None,
) -> str:
    """
    Abstracts a temporal profile model to a string.

    Parameters
    ----------------
    temporal_profile
        Temporal profile
    parameters
        Parameters of the method, including:
        - Parameters.INCLUDE_HEADER => includes the header in the response

    Returns
    ----------------
    text_abstr
        Textual abstraction of the log skeleton
    """
    if parameters is None:
        parameters = {}

    include_header = exec_utils.get_param_value(
        Parameters.INCLUDE_HEADER, parameters, True
    )

    ret = ["\n"]

    if include_header:
        ret.append(
            get_model_description() + "For this process, the model is:\n"
        )

    for act_couple, agg in temporal_profile.items():
        ret.append(
            "%s -> %s :  AVG: %.2f s  STD: %.2f s"
            % (act_couple[0], act_couple[1], agg[0], agg[1])
        )

    return "\n".join(ret)



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import Optional, Dict, Any, Union, List, Set, Tuple, Collection
import polars as pl
from collections import Counter
from pm4py.util import constants, xes_constants, exec_utils
from enum import Enum


class Parameters(Enum):
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY


def pandas_numpy_variants_apply_polars(
    lf: pl.LazyFrame, parameters=None
) -> Tuple[Dict[Tuple[str, ...], int], Dict[str, Tuple[str, ...]]]:
    """
    Efficient method returning the variants from a Polars LazyFrame

    Parameters
    ------------------
    lf
        LazyFrame
    parameters
        Parameters of the algorithm, including:
        - Parameters.CASE_ID_KEY => the case identifier
        - Parameters.ACTIVITY_KEY => the activity
        - Parameters.TIMESTAMP_KEY => the timestamp

    Returns
    ------------------
    variants_dict
        Dictionary associating to each variant the number of occurrences in the dataframe
    case_variant
        Dictionary associating to each case identifier the corresponding variant
    """
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )

    # Sort by case and timestamp to ensure proper order
    sorted_lf = lf.sort([case_id_key, timestamp_key])
    
    # Group by case and collect activities as lists
    # In Polars, aggregating without specifying a function collects into lists by default
    case_variants_df = (
        sorted_lf.group_by(case_id_key, maintain_order=True)
        .agg(pl.col(activity_key))
        .collect()
    )
    
    case_variant = {}
    variants_counter = Counter()
    
    for row in case_variants_df.iter_rows():
        case_id = row[0]
        activities = tuple(row[1])  # Convert list to tuple
        case_variant[case_id] = activities
        variants_counter[activities] += 1
    
    # Return as Python dictionaries
    variants_dict = {x: y for x, y in variants_counter.items()}
    
    return variants_dict, case_variant


def get_variants_count(
    lf: pl.LazyFrame, parameters: Optional[Dict[Any, Any]] = None
) -> Dict[Tuple[str, ...], int]:
    """
    Gets the dictionary of variants from the current LazyFrame

    Parameters
    --------------
    lf
        LazyFrame
    parameters
        Possible parameters of the algorithm, including:
            Parameters.ACTIVITY_KEY -> Column that contains the activity

    Returns
    --------------
    variants_dict
        Dictionary of variants in the log
    """
    if parameters is None:
        parameters = {}

    variants_counter, case_variant = pandas_numpy_variants_apply_polars(
        lf, parameters=parameters
    )

    return variants_counter


def get_variants_set(
    lf: pl.LazyFrame, parameters: Optional[Dict[Any, Any]] = None
) -> Set[Tuple[str, ...]]:
    """
    Gets the set of variants from the current LazyFrame

    Parameters
    --------------
    lf
        LazyFrame
    parameters
        Possible parameters of the algorithm, including:
            Parameters.ACTIVITY_KEY -> Column that contains the activity

    Returns
    --------------
    variants_set
        Set of variants in the log
    """
    if parameters is None:
        parameters = {}

    variants_dict = get_variants_count(lf, parameters=parameters)

    return set(variants_dict.keys())


'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from enum import Enum
from pm4py.objects.log.obj import EventLog
from pm4py.algo.discovery.declare.templates import *
import pandas as pd
from typing import Union, Dict, Optional, Any, Tuple, Collection, Set, List
from typing import Counter as TCounter
from pm4py.util import exec_utils, constants, xes_constants, pandas_utils
from collections import Counter
import numpy as np


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    CONSIDERED_ACTIVITIES = "considered_activities"
    MIN_SUPPORT_RATIO = "min_support_ratio"
    MIN_CONFIDENCE_RATIO = "min_confidence_ratio"
    AUTO_SELECTION_MULTIPLIER = "auto_selection_multiplier"
    ALLOWED_TEMPLATES = "allowed_templates"


def __rule_existence_column(act: str) -> Tuple[str, str]:
    return (EXISTENCE, act)


def __rule_exactly_one_column(act: str) -> Tuple[str, str]:
    return (EXACTLY_ONE, act)


def __rule_init_column(act: str) -> Tuple[str, str]:
    return (INIT, act)


def __rule_responded_existence_column(
    act: str, act2: str
) -> Tuple[str, str, str]:
    return (RESPONDED_EXISTENCE, act, act2)


def __rule_response(act: str, act2: str) -> Tuple[str, str, str]:
    return (RESPONSE, act, act2)


def __rule_precedence(act: str, act2: str) -> Tuple[str, str, str]:
    return (PRECEDENCE, act, act2)


def __rule_succession(act: str, act2: str) -> Tuple[str, str, str]:
    return (SUCCESSION, act, act2)


def __rule_alternate_response(act: str, act2: str) -> Tuple[str, str, str]:
    return (ALTRESPONSE, act, act2)


def __rule_alternate_precedence(act: str, act2: str) -> Tuple[str, str, str]:
    return (ALTPRECEDENCE, act, act2)


def __rule_alternate_succession(act: str, act2: str) -> Tuple[str, str, str]:
    return (ALTSUCCESSION, act, act2)


def __rule_chain_response(act: str, act2: str) -> Tuple[str, str, str]:
    return (CHAINRESPONSE, act, act2)


def __rule_chain_precedence(act: str, act2: str) -> Tuple[str, str, str]:
    return (CHAINPRECEDENCE, act, act2)


def __rule_chain_succession(act: str, act2: str) -> Tuple[str, str, str]:
    return (CHAINSUCCESSION, act, act2)


def __rule_absence_act(act: str) -> Tuple[str, str]:
    return (ABSENCE, act)


def __rule_coexistence(act: str, act2: str) -> Tuple[str, str, str]:
    return (COEXISTENCE, act, act2)


def __rule_non_coexistence(act: str, act2: str) -> Tuple[str, str, str]:
    return (NONCOEXISTENCE, act, act2)


def __rule_non_succession(act: str, act2: str) -> Tuple[str, str, str]:
    return (NONSUCCESSION, act, act2)


def __rule_non_chain_succession(act: str, act2: str) -> Tuple[str, str, str]:
    return (NONCHAINSUCCESSION, act, act2)


def __col_to_dict_rule(
    col_name: Union[Tuple[str, str], Tuple[str, str, str]]
) -> Tuple[str, Any]:
    if len(col_name) == 2:
        return col_name[0], col_name[1]
    else:
        if col_name[2] is None or pd.isna(col_name[2]) or not col_name[2]:
            return col_name[0], col_name[1]

        return col_name[0], (col_name[1], col_name[2])


def existence_template_step1(
    rules: Dict[Union[Tuple[str, str], Tuple[str, str, str]], int],
    trace: Collection[str],
    activities: Set[str],
    act_counter: TCounter[str],
    act_idxs: Dict[str, List[int]],
    allowed_templates: Collection[str],
):
    if EXISTENCE in allowed_templates:
        for act in activities:
            if act in act_counter:
                rules[__rule_existence_column(act)] = 1
            else:
                rules[__rule_existence_column(act)] = -1


def exactly_one_template_step1(
    rules: Dict[Union[Tuple[str, str], Tuple[str, str, str]], int],
    trace: Collection[str],
    activities: Set[str],
    act_counter: TCounter[str],
    act_idxs: Dict[str, List[int]],
    allowed_templates: Collection[str],
):
    if EXACTLY_ONE in allowed_templates:
        for act in activities:
            if act in act_counter:
                if act_counter[act] == 1:
                    rules[__rule_exactly_one_column(act)] = 1
                else:
                    rules[__rule_exactly_one_column(act)] = -1


def init_template_step1(
    rules: Dict[Union[Tuple[str, str], Tuple[str, str, str]], int],
    trace: Collection[str],
    activities: Set[str],
    act_counter: TCounter[str],
    act_idxs: Dict[str, List[int]],
    allowed_templates: Collection[str],
):
    if INIT in allowed_templates:
        for act in activities:
            if act == trace[0]:
                rules[__rule_init_column(act)] = 1
            else:
                rules[__rule_init_column(act)] = -1


def responded_existence_template_step1(
    rules: Dict[Union[Tuple[str, str], Tuple[str, str, str]], int],
    trace: Collection[str],
    activities: Set[str],
    act_counter: TCounter[str],
    act_idxs: Dict[str, List[int]],
    allowed_templates: Collection[str],
):
    if RESPONDED_EXISTENCE in allowed_templates:
        for act in act_counter:
            for act2 in activities:
                if act2 != act:
                    if act2 not in act_counter:
                        rules[__rule_responded_existence_column(act, act2)] = (
                            -1
                        )
                    else:
                        rules[__rule_responded_existence_column(act, act2)] = 1


def response_template_step1(
    rules: Dict[Union[Tuple[str, str], Tuple[str, str, str]], int],
    trace: Collection[str],
    activities: Set[str],
    act_counter: TCounter[str],
    act_idxs: Dict[str, List[int]],
    allowed_templates: Collection[str],
):
    if RESPONSE in allowed_templates:
        for act in act_counter:
            for act2 in activities:
                if act2 != act:
                    if act2 not in act_counter:
                        rules[__rule_response(act, act2)] = -1
                    else:
                        if act_idxs[act][-1] < act_idxs[act2][-1]:
                            rules[__rule_response(act, act2)] = 1
                        else:
                            rules[__rule_response(act, act2)] = -1


def precedence_template_step1(
    rules: Dict[Union[Tuple[str, str], Tuple[str, str, str]], int],
    trace: Collection[str],
    activities: Set[str],
    act_counter: TCounter[str],
    act_idxs: Dict[str, List[int]],
    allowed_templates: Collection[str],
):
    if PRECEDENCE in allowed_templates:
        for act in act_counter:
            for act2 in activities:
                if act2 != act:
                    if act2 not in act_counter:
                        pass
                    else:
                        if act_idxs[act2][0] < act_idxs[act][0]:
                            rules[__rule_precedence(act2, act)] = 1
                        else:
                            rules[__rule_precedence(act2, act)] = -1


def altresponse_template_step1(
    rules: Dict[Union[Tuple[str, str], Tuple[str, str, str]], int],
    trace: Collection[str],
    activities: Set[str],
    act_counter: TCounter[str],
    act_idxs: Dict[str, List[int]],
    allowed_templates: Collection[str],
):
    if ALTRESPONSE in allowed_templates:
        for act in act_counter:
            for act2 in activities:
                if act2 != act:
                    if act2 not in act_counter:
                        rules[__rule_alternate_response(act, act2)] = -1
                    else:
                        is_ok_alt_resp = False
                        if len(act_idxs[act]) == len(act_idxs[act2]):
                            lenn = len(act_idxs[act])
                            is_ok_alt_resp = True

                            for i in range(lenn):
                                if act_idxs[act][i] > act_idxs[act2][i] or (
                                    i < lenn - 1
                                    and act_idxs[act][i + 1]
                                    < act_idxs[act2][i]
                                ):
                                    is_ok_alt_resp = False
                                    break
                                elif act_idxs[act][i] + 1 != act_idxs[act2][i]:
                                    pass
                        rules[__rule_alternate_response(act, act2)] = (
                            1 if is_ok_alt_resp else -1
                        )


def chainresponse_template_step1(
    rules: Dict[Union[Tuple[str, str], Tuple[str, str, str]], int],
    trace: Collection[str],
    activities: Set[str],
    act_counter: TCounter[str],
    act_idxs: Dict[str, List[int]],
    allowed_templates: Collection[str],
):
    if CHAINRESPONSE in allowed_templates:
        for act in act_counter:
            for act2 in activities:
                if act2 != act:
                    if act2 not in act_counter:
                        rules[__rule_chain_response(act, act2)] = -1
                    else:
                        is_ok_chain_resp = False
                        if len(act_idxs[act]) == len(act_idxs[act2]):
                            lenn = len(act_idxs[act])
                            is_ok_chain_resp = True

                            for i in range(lenn):
                                if act_idxs[act][i] > act_idxs[act2][i] or (
                                    i < lenn - 1
                                    and act_idxs[act][i + 1]
                                    < act_idxs[act2][i]
                                ):
                                    is_ok_chain_resp = False
                                    break
                                elif act_idxs[act][i] + 1 != act_idxs[act2][i]:
                                    is_ok_chain_resp = False
                                    break
                        rules[__rule_chain_response(act, act2)] = (
                            1 if is_ok_chain_resp else -1
                        )


def altprecedence_template_step1(
    rules: Dict[Union[Tuple[str, str], Tuple[str, str, str]], int],
    trace: Collection[str],
    activities: Set[str],
    act_counter: TCounter[str],
    act_idxs: Dict[str, List[int]],
    allowed_templates: Collection[str],
):
    if ALTPRECEDENCE in allowed_templates:
        for act in act_counter:
            for act2 in activities:
                if act2 != act:
                    if act2 not in act_counter:
                        pass
                    else:
                        is_ok_alt_prec = False
                        if len(act_idxs[act]) == len(act_idxs[act2]):
                            lenn = len(act_idxs[act])
                            is_ok_alt_prec = True
                            for i in range(lenn):
                                if act_idxs[act2][i] > act_idxs[act][i] or (
                                    i < lenn - 1
                                    and act_idxs[act2][i + 1]
                                    < act_idxs[act][i]
                                ):
                                    is_ok_alt_prec = False
                                    break
                                elif act_idxs[act2][i] + 1 != act_idxs[act][i]:
                                    pass
                        rules[__rule_alternate_precedence(act2, act)] = (
                            1 if is_ok_alt_prec else -1
                        )


def chainprecedence_template_step1(
    rules: Dict[Union[Tuple[str, str], Tuple[str, str, str]], int],
    trace: Collection[str],
    activities: Set[str],
    act_counter: TCounter[str],
    act_idxs: Dict[str, List[int]],
    allowed_templates: Collection[str],
):
    if CHAINPRECEDENCE in allowed_templates:
        for act in act_counter:
            for act2 in activities:
                if act2 != act:
                    if act2 not in act_counter:
                        pass
                    else:
                        is_ok_chain_prec = False
                        if len(act_idxs[act]) == len(act_idxs[act2]):
                            lenn = len(act_idxs[act])
                            is_ok_chain_prec = True
                            # check alternate and chain response
                            for i in range(lenn):
                                if act_idxs[act2][i] > act_idxs[act][i] or (
                                    i < lenn - 1
                                    and act_idxs[act2][i + 1]
                                    < act_idxs[act][i]
                                ):
                                    is_ok_chain_prec = False
                                    break
                                elif act_idxs[act2][i] + 1 != act_idxs[act][i]:
                                    is_ok_chain_prec = False
                                    break
                        rules[__rule_chain_precedence(act2, act)] = (
                            1 if is_ok_chain_prec else -1
                        )


def absence_template(
    table,
    columns: Collection[str],
    activities: Set[str],
    allowed_templates: Collection[str],
) -> pd.DataFrame:
    if ABSENCE in allowed_templates and EXISTENCE in allowed_templates:
        for act in activities:
            table[__rule_absence_act(act)] = (
                -1 * table[__rule_existence_column(act)]
            )
    return table


def exactly_one_template_step2(
    table,
    columns: Collection[str],
    activities: Set[str],
    allowed_templates: Collection[str],
) -> pd.DataFrame:
    if EXACTLY_ONE in allowed_templates:
        for act in activities:
            if __rule_exactly_one_column(act) not in columns:
                table[__rule_exactly_one_column(act)] = [0] * len(
                    table[list(table.keys())[0]]
                )
    return table


def responded_existence_template_step2(
    table,
    columns: Collection[str],
    activities: Set[str],
    allowed_templates: Collection[str],
) -> pd.DataFrame:
    if RESPONDED_EXISTENCE in allowed_templates:
        for act in activities:
            for act2 in activities:
                if act2 != act:
                    if (
                        __rule_responded_existence_column(act, act2)
                        not in columns
                    ):
                        table[__rule_responded_existence_column(act, act2)] = [
                            0
                        ] * len(table[list(table.keys())[0]])
    return table


def response_template_step2(
    table,
    columns: Collection[str],
    activities: Set[str],
    allowed_templates: Collection[str],
) -> pd.DataFrame:
    if RESPONSE in allowed_templates:
        for act in activities:
            for act2 in activities:
                if act2 != act:
                    if __rule_response(act, act2) not in columns:
                        table[__rule_response(act, act2)] = [0] * len(
                            table[list(table.keys())[0]]
                        )
    return table


def precedence_template_step2(
    table,
    columns: Collection[str],
    activities: Set[str],
    allowed_templates: Collection[str],
) -> pd.DataFrame:
    if PRECEDENCE in allowed_templates:
        for act in activities:
            for act2 in activities:
                if act2 != act:
                    if __rule_precedence(act, act2) not in columns:
                        table[__rule_precedence(act, act2)] = [0] * len(
                            table[list(table.keys())[0]]
                        )
    return table


def altresponse_template_step2(
    table,
    columns: Collection[str],
    activities: Set[str],
    allowed_templates: Collection[str],
) -> pd.DataFrame:
    if ALTRESPONSE in allowed_templates:
        for act in activities:
            for act2 in activities:
                if act2 != act:
                    if __rule_alternate_response(act, act2) not in columns:
                        table[__rule_alternate_response(act, act2)] = [
                            0
                        ] * len(table[list(table.keys())[0]])
    return table


def chainresponse_template_step2(
    table,
    columns: Collection[str],
    activities: Set[str],
    allowed_templates: Collection[str],
) -> pd.DataFrame:
    if CHAINRESPONSE in allowed_templates:
        for act in activities:
            for act2 in activities:
                if act2 != act:
                    if __rule_chain_response(act, act2) not in columns:
                        table[__rule_chain_response(act, act2)] = [0] * len(
                            table[list(table.keys())[0]]
                        )
    return table


def altprecedence_template_step2(
    table,
    columns: Collection[str],
    activities: Set[str],
    allowed_templates: Collection[str],
) -> pd.DataFrame:
    if ALTPRECEDENCE in allowed_templates:
        for act in activities:
            for act2 in activities:
                if act2 != act:
                    if __rule_alternate_precedence(act, act2) not in columns:
                        table[__rule_alternate_precedence(act, act2)] = [
                            0
                        ] * len(table[list(table.keys())[0]])
    return table


def chainprecedence_template_step2(
    table,
    columns: Collection[str],
    activities: Set[str],
    allowed_templates: Collection[str],
) -> pd.DataFrame:
    if CHAINPRECEDENCE in allowed_templates:
        for act in activities:
            for act2 in activities:
                if act2 != act:
                    if __rule_chain_precedence(act, act2) not in columns:
                        table[__rule_chain_precedence(act, act2)] = [0] * len(
                            table[list(table.keys())[0]]
                        )
    return table


def succession_template(
    table,
    columns: Collection[str],
    activities: Set[str],
    allowed_templates: Collection[str],
) -> pd.DataFrame:
    if (
        SUCCESSION in allowed_templates
        and RESPONSE in allowed_templates
        and PRECEDENCE in allowed_templates
    ):
        for act in activities:
            for act2 in activities:
                if act2 != act:
                    table[__rule_succession(act, act2)] = np.minimum(
                        table[__rule_response(act, act2)],
                        table[__rule_precedence(act, act2)],
                    )
    return table


def altsuccession_template(
    table,
    columns: Collection[str],
    activities: Set[str],
    allowed_templates: Collection[str],
) -> pd.DataFrame:
    if (
        ALTSUCCESSION in allowed_templates
        and ALTRESPONSE in allowed_templates
        and ALTPRECEDENCE in allowed_templates
    ):
        for act in activities:
            for act2 in activities:
                if act2 != act:
                    table[__rule_alternate_succession(act, act2)] = np.minimum(
                        table[__rule_alternate_response(act, act2)],
                        table[__rule_alternate_precedence(act, act2)],
                    )
    return table


def chainsuccession_template(
    table,
    columns: Collection[str],
    activities: Set[str],
    allowed_templates: Collection[str],
) -> pd.DataFrame:
    if (
        CHAINSUCCESSION in allowed_templates
        and CHAINRESPONSE in allowed_templates
        and CHAINPRECEDENCE in allowed_templates
    ):
        for act in activities:
            for act2 in activities:
                if act2 != act:
                    table[__rule_chain_succession(act, act2)] = np.minimum(
                        table[__rule_chain_response(act, act2)],
                        table[__rule_chain_precedence(act, act2)],
                    )
    return table


def coexistence_template(
    table,
    columns: Collection[str],
    activities: Set[str],
    allowed_templates: Collection[str],
) -> pd.DataFrame:
    if (
        COEXISTENCE in allowed_templates
        and RESPONDED_EXISTENCE in allowed_templates
    ):
        for act in activities:
            for act2 in activities:
                if act2 != act:
                    table[__rule_coexistence(act, act2)] = np.minimum(
                        table[__rule_responded_existence_column(act, act2)],
                        table[__rule_responded_existence_column(act2, act)],
                    )
    return table


def noncoexistence_template(
    table,
    columns: Collection[str],
    activities: Set[str],
    allowed_templates: Collection[str],
) -> pd.DataFrame:
    if (
        NONCOEXISTENCE in allowed_templates
        and COEXISTENCE in allowed_templates
        and RESPONDED_EXISTENCE in allowed_templates
    ):
        for act in activities:
            for act2 in activities:
                if act2 != act:
                    table[__rule_non_coexistence(act, act2)] = (
                        -1 * table[__rule_coexistence(act, act2)]
                    )
    return table


def nonsuccession_template(
    table,
    columns: Collection[str],
    activities: Set[str],
    allowed_templates: Collection[str],
) -> pd.DataFrame:
    if (
        NONSUCCESSION in allowed_templates
        and SUCCESSION in allowed_templates
        and RESPONSE in allowed_templates
        and PRECEDENCE in allowed_templates
    ):
        for act in activities:
            for act2 in activities:
                if act2 != act:
                    table[__rule_non_succession(act, act2)] = (
                        -1 * table[__rule_succession(act, act2)]
                    )
    return table


def nonchainsuccession_template(
    table,
    columns: Collection[str],
    activities: Set[str],
    allowed_templates: Collection[str],
) -> pd.DataFrame:
    if (
        NONCHAINSUCCESSION in allowed_templates
        and CHAINSUCCESSION in allowed_templates
        and CHAINRESPONSE in allowed_templates
        and CHAINPRECEDENCE in allowed_templates
    ):
        for act in activities:
            for act2 in activities:
                if act2 != act:
                    table[__rule_non_chain_succession(act, act2)] = (
                        -1 * table[__rule_chain_succession(act, act2)]
                    )
    return table


def form_rules_table(
    log: Union[EventLog, pd.DataFrame],
    parameters: Optional[Dict[Any, Any]] = None,
) -> pd.DataFrame:
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )

    allowed_templates = exec_utils.get_param_value(
        Parameters.ALLOWED_TEMPLATES, parameters, None
    )

    if allowed_templates is None:
        allowed_templates = {
            EXISTENCE,
            EXACTLY_ONE,
            INIT,
            RESPONDED_EXISTENCE,
            RESPONSE,
            PRECEDENCE,
            SUCCESSION,
            ALTRESPONSE,
            ALTPRECEDENCE,
            ALTSUCCESSION,
            CHAINRESPONSE,
            CHAINPRECEDENCE,
            CHAINSUCCESSION,
            ABSENCE,
            COEXISTENCE,
        }

    import pm4py

    projected_log = pm4py.project_on_event_attribute(
        log, activity_key, case_id_key=case_id_key
    )
    activities = exec_utils.get_param_value(
        Parameters.CONSIDERED_ACTIVITIES, parameters, None
    )

    if activities is None:
        activities = set(y for x in projected_log for y in x)

    allowed_templates = set(allowed_templates)
    activities = set(activities)

    vars = Counter(
        [tuple([y for y in x if y in activities]) for x in projected_log]
    )
    table = []

    for trace, occs in vars.items():
        act_counter = Counter(trace)
        act_idxs = {x: [] for x in set(trace)}

        for idx, act in enumerate(trace):
            act_idxs[act].append(idx)

        rules = {}

        existence_template_step1(
            rules, trace, activities, act_counter, act_idxs, allowed_templates
        )
        exactly_one_template_step1(
            rules, trace, activities, act_counter, act_idxs, allowed_templates
        )
        init_template_step1(
            rules, trace, activities, act_counter, act_idxs, allowed_templates
        )
        responded_existence_template_step1(
            rules, trace, activities, act_counter, act_idxs, allowed_templates
        )
        response_template_step1(
            rules, trace, activities, act_counter, act_idxs, allowed_templates
        )
        precedence_template_step1(
            rules, trace, activities, act_counter, act_idxs, allowed_templates
        )
        altresponse_template_step1(
            rules, trace, activities, act_counter, act_idxs, allowed_templates
        )
        chainresponse_template_step1(
            rules, trace, activities, act_counter, act_idxs, allowed_templates
        )
        altprecedence_template_step1(
            rules, trace, activities, act_counter, act_idxs, allowed_templates
        )
        chainprecedence_template_step1(
            rules, trace, activities, act_counter, act_idxs, allowed_templates
        )

        for i in range(occs):
            table.append(rules)

    columns = set(y for x in table for y in x.keys())
    table2 = {c: [] for c in columns}
    for i in range(len(table)):
        trace = table[i]
        for c in columns:
            if c in trace:
                table2[c].append(trace[c])
            else:
                table2[c].append(0)

    for c in table2:
        table2[c] = np.array(table2[c])

    table2 = absence_template(table2, columns, activities, allowed_templates)
    table2 = exactly_one_template_step2(
        table2, columns, activities, allowed_templates
    )
    table2 = responded_existence_template_step2(
        table2, columns, activities, allowed_templates
    )
    table2 = response_template_step2(
        table2, columns, activities, allowed_templates
    )
    table2 = precedence_template_step2(
        table2, columns, activities, allowed_templates
    )
    table2 = altresponse_template_step2(
        table2, columns, activities, allowed_templates
    )
    table2 = chainresponse_template_step2(
        table2, columns, activities, allowed_templates
    )
    table2 = altprecedence_template_step2(
        table2, columns, activities, allowed_templates
    )
    table2 = chainprecedence_template_step2(
        table2, columns, activities, allowed_templates
    )

    table2 = succession_template(
        table2, columns, activities, allowed_templates
    )
    table2 = altsuccession_template(
        table2, columns, activities, allowed_templates
    )
    table2 = chainsuccession_template(
        table2, columns, activities, allowed_templates
    )
    table2 = coexistence_template(
        table2, columns, activities, allowed_templates
    )
    table2 = noncoexistence_template(
        table2, columns, activities, allowed_templates
    )
    table2 = nonsuccession_template(
        table2, columns, activities, allowed_templates
    )
    table2 = nonchainsuccession_template(
        table2, columns, activities, allowed_templates
    )

    return pandas_utils.instantiate_dataframe(table2)


def get_rules_from_rules_df(
    rules_df, parameters: Optional[Dict[Any, Any]] = None
) -> Dict[str, Dict[Any, Dict[str, int]]]:
    if parameters is None:
        parameters = {}

    min_support_ratio = exec_utils.get_param_value(
        Parameters.MIN_SUPPORT_RATIO, parameters, None
    )
    min_confidence_ratio = exec_utils.get_param_value(
        Parameters.MIN_CONFIDENCE_RATIO, parameters, None
    )
    rules = {}

    if min_support_ratio is None and min_confidence_ratio is None:
        # auto determine the minimum support and confidence ratio by
        # identifying the values for the best feature
        auto_selection_multiplier = exec_utils.get_param_value(
            Parameters.AUTO_SELECTION_MULTIPLIER, parameters, 0.8
        )
        cols_prod = []
        for col_name in rules_df:
            col = rules_df[col_name]
            supp = len(col[col != 0])
            supp_ratio = float(supp) / float(len(rules_df))
            if supp_ratio > 0:
                conf_ratio = float(len(col[col == 1])) / float(supp)
                prod = supp_ratio * conf_ratio
                cols_prod.append((col_name, prod))
        cols_prod = sorted(cols_prod, key=lambda x: (x[1], x[0]), reverse=True)
        col = rules_df[cols_prod[0][0]]
        supp = len(col[col != 0])
        min_support_ratio = (
            float(supp) / float(len(rules_df)) * auto_selection_multiplier
        )
        min_confidence_ratio = (
            float(len(col[col == 1])) / float(supp) * auto_selection_multiplier
        )

    if rules_df is not None and len(rules_df) > 0:
        for col_name in rules_df:
            col = rules_df[col_name]
            supp = len(col[col != 0])

            if supp >= len(rules_df) * min_support_ratio:
                conf = len(col[col == 1])

                if conf >= supp * min_confidence_ratio:
                    rule, key = __col_to_dict_rule(col_name)
                    if rule not in rules:
                        rules[rule] = {}

                    rules[rule][key] = {"support": supp, "confidence": conf}

    return rules


def apply(
    log: Union[EventLog, pd.DataFrame],
    parameters: Optional[Dict[Any, Any]] = None,
) -> Dict[str, Dict[Any, Dict[str, int]]]:
    """
    Discovers a DECLARE model from the provided event log

    Paper:
    F. M. Maggi, A. J. Mooij and W. M. P. van der Aalst, "User-guided discovery of declarative process models," 2011 IEEE Symposium on Computational Intelligence and Data Mining (CIDM), Paris, France, 2011, pp. 192-199, doi: 10.1109/CIDM.2011.5949297.


    Parameters
    ---------------
    log
        Log object (EventLog, Pandas table)
    parameters
        Possible parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY
        - Parameters.CONSIDERED_ACTIVITIES
        - Parameters.MIN_SUPPORT_RATIO
        - Parameters.MIN_CONFIDENCE_RATIO
        - Parameters.AUTO_SELECTION_MULTIPLIER
        - Parameters.ALLOWED_TEMPLATES: collection of templates to consider, including:
            * existence
            * exactly_one
            * init
            * responded_existence
            * response
            * precedence
            * succession
            * altresponse
            * altprecedence
            * altsuccession
            * chainresponse
            * chainprecedence
            * chainsuccession
            * absence
            * coexistence
            * noncoexistence
            * nonsuccession
            * nonchainsuccession

    Returns
    -------------
    declare_model
        DECLARE model (as Python dictionary), where each template is associated with its own rules
    """
    if parameters is None:
        parameters = {}

    rules_df = form_rules_table(log, parameters=parameters)

    rules = get_rules_from_rules_df(rules_df, parameters=parameters)

    return rules



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum

from pm4py.objects.conversion.log import converter as log_converter
from pm4py.util import exec_utils
from pm4py.util.constants import (
    PARAMETER_CONSTANT_ATTRIBUTE_KEY,
    PARAMETER_CONSTANT_RESOURCE_KEY,
    PARAMETER_CONSTANT_TIMESTAMP_KEY,
)
from pm4py.util.xes_constants import (
    DEFAULT_NAME_KEY,
    DEFAULT_RESOURCE_KEY,
    DEFAULT_TIMESTAMP_KEY,
)

from typing import Optional, Dict, Any, Union, List
from pm4py.objects.log.obj import EventLog
import itertools


class Parameters(Enum):
    ATTRIBUTE_KEY = PARAMETER_CONSTANT_ATTRIBUTE_KEY
    TIMESTAMP_KEY = PARAMETER_CONSTANT_TIMESTAMP_KEY
    RESOURCE_KEY = PARAMETER_CONSTANT_RESOURCE_KEY
    POSITIVE = "positive"
    ENABLE_TIMESTAMP = "enable_timestamp"
    TIMESTAMP_DIFF_BOUNDARIES = "timestamp_diff_boundaries"


POSITIVE = Parameters.POSITIVE
ENABLE_TIMESTAMP = Parameters.ENABLE_TIMESTAMP
TIMESTAMP_DIFF_BOUNDARIES = Parameters.TIMESTAMP_DIFF_BOUNDARIES


def timestamp_list_is_ge(a, b):
    for i in range(len(a)):
        if a[i] < b[i][0]:
            return False
    return True


def timestamp_list_is_le(a, b):
    for i in range(len(a)):
        if a[i] > b[i][1]:
            return False
    return True


def eventually_follows(
    log: EventLog,
    attribute_values: List[str],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> EventLog:
    """
    Applies the eventually follows rule

    Parameters
    ------------
    log
        Log
    attribute_values
        A list of attribute_values attribute_values[n] follows attribute_values[n-1] follows ... follows attribute_values[0]

    parameters
        Parameters of the algorithm, including the attribute key and the positive parameter:
        - If True, returns all the cases containing all attribute_values and in which attribute_values[i] was eventually followed by attribute_values[i + 1]
        - If False, returns all the cases not containing all attribute_values, or in which an instance of attribute_values[i] was not eventually
        followed by an instance of attribute_values[i + 1]

    Returns
    ------------
    filtered_log
        Filtered log
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )

    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, DEFAULT_TIMESTAMP_KEY
    )

    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )
    enable_timestamp = exec_utils.get_param_value(
        Parameters.ENABLE_TIMESTAMP, parameters, False
    )
    timestamp_diff_boundaries = exec_utils.get_param_value(
        Parameters.TIMESTAMP_DIFF_BOUNDARIES, parameters, []
    )

    new_log = EventLog(
        list(),
        attributes=log.attributes,
        extensions=log.extensions,
        classifiers=log.classifiers,
        omni_present=log.omni_present,
        properties=log.properties,
    )

    for trace in log:
        occurrences = [
            [
                i
                for i in range(len(trace))
                if attribute_key in trace[i]
                and trace[i][attribute_key] == attribute_value
            ]
            for attribute_value in attribute_values
        ]

        is_good = False

        for c in itertools.product(*occurrences):
            ok = True
            for i in range(len(c) - 1):
                if c[i] >= c[i + 1]:
                    ok = False
                    break
            if ok:
                if enable_timestamp and timestamp_diff_boundaries:
                    for i in range(len(c) - 1):
                        timest_i = trace[i][timestamp_key].timestamp()
                        timest_j = trace[i + 1][timestamp_key].timestamp()
                        if (
                            timest_j - timest_i
                            < timestamp_diff_boundaries[i][0]
                            or timest_j - timest_i
                            > timestamp_diff_boundaries[i][1]
                        ):
                            ok = False
                            break

            if ok:
                is_good = True
                break

        if is_good:
            if positive:
                new_log.append(trace)
        elif not positive:
            new_log.append(trace)

    return new_log


def A_next_B_next_C(
    log: EventLog,
    A: str,
    B: str,
    C: str,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> EventLog:
    """
    Applies the A next B next C rule

    Parameters
    ------------
    log
        Log
    A
        A attribute value
    B
        B attribute value
    C
        C attribute value
    parameters
        Parameters of the algorithm, including the attribute key and the positive parameter:
        - If True, returns all the cases containing A, B and C and in which A was directly followed by B and B was directly followed by C
        - If False, returns all the cases not containing A or B or C, or in which none instance of A was directly
        followed by an instance of B and B was directly followed by C

    Returns
    ------------
    filtered_log
        Filtered log
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )

    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )

    new_log = EventLog(
        list(),
        attributes=log.attributes,
        extensions=log.extensions,
        classifiers=log.classifiers,
        omni_present=log.omni_present,
        properties=log.properties,
    )

    for trace in log:
        occ_A = [
            i
            for i in range(len(trace))
            if attribute_key in trace[i] and trace[i][attribute_key] == A
        ]
        occ_B = [
            i
            for i in range(len(trace))
            if attribute_key in trace[i] and trace[i][attribute_key] == B
        ]
        occ_C = [
            i
            for i in range(len(trace))
            if attribute_key in trace[i] and trace[i][attribute_key] == C
        ]

        found = False

        for a in occ_A:
            for b in occ_B:
                for c in occ_C:
                    if (b - a) == 1 and (c - b) == 1:
                        found = True

        if found:
            if positive:
                new_log.append(trace)
        elif not positive:
            new_log.append(trace)

    return new_log


def four_eyes_principle(
    log: EventLog,
    A: str,
    B: str,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> EventLog:
    """
    Verifies the Four Eyes Principle given A and B

    Parameters
    -------------
    log
        Log
    A
        A attribute value
    B
        B attribute value
    parameters
        Parameters of the algorithm, including the attribute key and the positive parameter:
        - if True, then filters all the cases containing A and B which have empty intersection between the set
          of resources doing A and B
        - if False, then filters all the cases containing A and B which have no empty intersection between the set
          of resources doing A and B

    Returns
    --------------
    filtered_log
        Filtered log
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )

    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )
    resource_key = exec_utils.get_param_value(
        Parameters.RESOURCE_KEY, parameters, DEFAULT_RESOURCE_KEY
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )

    new_log = EventLog(
        list(),
        attributes=log.attributes,
        extensions=log.extensions,
        classifiers=log.classifiers,
        omni_present=log.omni_present,
        properties=log.properties,
    )

    for trace in log:
        occ_A = set(
            [
                trace[i][resource_key]
                for i in range(len(trace))
                if attribute_key in trace[i]
                and resource_key in trace[i]
                and trace[i][attribute_key] == A
            ]
        )
        occ_B = set(
            [
                trace[i][resource_key]
                for i in range(len(trace))
                if attribute_key in trace[i]
                and resource_key in trace[i]
                and trace[i][attribute_key] == B
            ]
        )

        if len(occ_A) > 0 and len(occ_B) > 0:
            inte = occ_A.intersection(occ_B)

            if not positive and len(inte) > 0:
                new_log.append(trace)
            elif positive and len(inte) == 0:
                new_log.append(trace)

    return new_log


def attr_value_different_persons(
    log: EventLog,
    A: str,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> EventLog:
    """
    Checks whether an attribute value is assumed on events done by different resources

    Parameters
    ------------
    log
        Log
    A
        A attribute value
    parameters
        Parameters of the algorithm, including the attribute key and the positive parameter:
            - if True, then filters all the cases containing occurrences of A done by different resources
            - if False, then filters all the cases not containing occurrences of A done by different resources

    Returns
    -------------
    filtered_log
        Filtered log
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )

    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )
    resource_key = exec_utils.get_param_value(
        Parameters.RESOURCE_KEY, parameters, DEFAULT_RESOURCE_KEY
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )

    new_log = EventLog(
        list(),
        attributes=log.attributes,
        extensions=log.extensions,
        classifiers=log.classifiers,
        omni_present=log.omni_present,
        properties=log.properties,
    )

    for trace in log:
        occ_A = set(
            [
                trace[i][resource_key]
                for i in range(len(trace))
                if attribute_key in trace[i]
                and resource_key in trace[i]
                and trace[i][attribute_key] == A
            ]
        )
        if len(occ_A) > 1:
            if positive:
                new_log.append(trace)
        elif not positive:
            new_log.append(trace)

    return new_log



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.discovery.ocel.interleavings.variants import (
    timestamp_interleavings,
)
from enum import Enum
from pm4py.util import exec_utils
import pandas as pd
from typing import Optional, Dict, Any


class Variants(Enum):
    TIMESTAMP_INTERLEAVINGS = timestamp_interleavings


def apply(
    left_df: pd.DataFrame,
    right_df: pd.DataFrame,
    case_relations: pd.DataFrame,
    variant=Variants.TIMESTAMP_INTERLEAVINGS,
    parameters: Optional[Dict[Any, Any]] = None,
):
    """
    Discover the interleavings between two dataframes, given also a dataframe about the relations of the cases.

    Parameters
    -----------------
    left_df
        Left dataframe
    right_df
        Right dataframe
    case_relations
        Dictionary associating the cases of the first dataframe (column: case:concept:name_LEFT) to the
        cases of the second dataframe (column: case:concept:name_RIGHT)
    variant
        Variant of the algorithm to be used, possible values:
        - Variants.TIMESTAMP_INTERLEAVINGS
    parameters
        Variant-specific parameters

    Returns
    -----------------
    interleavings
        Interleavings dataframe
    """
    return exec_utils.get_variant(variant).apply(
        left_df, right_df, case_relations, parameters
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util import nx_utils
import sys


def exact_match_minimum_average(l1, l2):
    """
    Performs an exact matching, having minimum average,
    of two lists of times (a timestamp of the first list
    is matched with a timestamp of the second list that
    is greater or equal). Some timestamps in the two lists
    may be left out of the matching.

    Parameters
    ---------------
    l1
        First list of times
    l2
        Second list of times

    Returns
    ---------------
    exact_matching
        Exact matching (list of tuples having as first element
        an element of the first list, and as second element
        an element of the second list)
    """
    G = nx_utils.Graph()
    for i in range(len(l1)):
        G.add_node(i)
    for j in range(len(l2)):
        G.add_node(len(l1) + j)
    for i in range(len(l1)):
        for j in range(len(l2)):
            if l1[i] <= l2[j]:
                G.add_edge(i, len(l1) + j, weight=(l2[j] - l1[i]))
            else:
                G.add_edge(i, len(l1) + j, weight=sys.maxsize)
    matching0 = {
        x: y
        for x, y in nx_utils.minimum_weight_full_matching(G).items()
        if x < len(l1)
    }
    matching = []
    for k1, k2 in matching0.items():
        v1 = l1[k1]
        v2 = l2[k2 - len(l1)]
        if v2 >= v1:
            matching.append((v1, v2))
    return matching



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Dict, Optional, Any, List, Union

import pandas as pd

from pm4py.statistics.overlap.utils import compute
from pm4py.util import exec_utils, constants, xes_constants


class Parameters(Enum):
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY


def apply(
    df: pd.DataFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> List[int]:
    """
    Computes the case overlap statistic from a Pandas dataframe

    Parameters
    -----------------
    df
        Dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.TIMESTAMP_KEY => attribute representing the completion timestamp
        - Parameters.START_TIMESTAMP_KEY => attribute representing the start timestamp

    Returns
    ----------------
    case_overlap
        List associating to each case the number of open cases during the life of a case
    """
    if parameters is None:
        parameters = {}

    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )

    columns = list({timestamp_key, start_timestamp_key, case_id_key})
    stream = df[columns].to_dict("records")

    points = []
    cases = []
    cases_points = {}
    for event in stream:
        case_id = event[case_id_key]
        if case_id not in cases:
            cases.append(case_id)
            cases_points[case_id] = []
        cases_points[case_id].append(
            (
                event[start_timestamp_key].timestamp(),
                event[timestamp_key].timestamp(),
            )
        )

    for case in cases:
        case_points = cases_points[case]
        points.append(
            (min(x[0] for x in case_points), max(x[1] for x in case_points))
        )

    return compute.apply(points, parameters=parameters)
